#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject
from datarobot.utils import datetime_to_string, from_api, parse_time

from ..enums import (
    DATA_SUBSET,
    DATETIME_TREND_PLOTS_RESOLUTION,
    DATETIME_TREND_PLOTS_STATUS,
    SOURCE_TYPE,
)
from ..utils import underscorize

_metadata_start_end_dates_trafaret = t.Dict(
    {
        t.Key("start_date"): t.Or(t.Null, t.Call(parse_time)),
        t.Key("end_date"): t.Or(t.Null, t.Call(parse_time)),
    }
).ignore_extra("*")

_backtest_holdout_metadata_trafaret = t.Dict(
    {
        t.Key(SOURCE_TYPE.TRAINING): _metadata_start_end_dates_trafaret,
        t.Key(SOURCE_TYPE.VALIDATION): _metadata_start_end_dates_trafaret,
    }
).ignore_extra("*")

_accuracy_anomaly_over_time_plots_status_trafaret = t.Dict(
    {
        t.Key(SOURCE_TYPE.TRAINING): t.Enum(*DATETIME_TREND_PLOTS_STATUS.ALL),
        t.Key(SOURCE_TYPE.VALIDATION): t.Enum(*DATETIME_TREND_PLOTS_STATUS.ALL),
    }
).ignore_extra("*")

_forecast_vs_actual_training_validation_status_trafaret = t.Dict(
    {
        t.Key(underscorize(status), to_name=status, optional=True): t.List(Int)
        for status in DATETIME_TREND_PLOTS_STATUS.ALL
    }
).ignore_extra("*")

_forecast_vs_actual_status_trafaret = t.Dict(
    {
        t.Key(SOURCE_TYPE.TRAINING): _forecast_vs_actual_training_validation_status_trafaret,
        t.Key(SOURCE_TYPE.VALIDATION): _forecast_vs_actual_training_validation_status_trafaret,
    }
).ignore_extra("*")

_accuracy_over_time_plots_bin_trafaret = t.Dict(
    {
        t.Key("start_date"): parse_time,
        t.Key("end_date"): parse_time,
        t.Key("actual"): t.Or(t.Float, t.Null),
        t.Key("predicted"): t.Or(t.Float, t.Null),
        t.Key("frequency"): t.Or(Int, t.Null),
    }
).ignore_extra("*")

_forecast_vs_actual_plots_bin_trafaret = t.Dict(
    {
        t.Key("start_date"): parse_time,
        t.Key("end_date"): parse_time,
        t.Key("actual"): t.Or(t.Float, t.Null),
        t.Key("error"): t.Or(t.Float, t.Null),
        t.Key("normalized_error"): t.Or(t.Float, t.Null),
        t.Key("forecasts"): t.List(t.Or(t.Float, t.Null)),
        t.Key("frequency"): t.Or(Int, t.Null),
    }
).ignore_extra("*")

_anomaly_over_time_plots_bin_trafaret = t.Dict(
    {
        t.Key("start_date"): parse_time,
        t.Key("end_date"): parse_time,
        t.Key("predicted"): t.Or(t.Float, t.Null),
        t.Key("frequency"): t.Or(Int, t.Null),
    }
).ignore_extra("*")

_datetime_trend_plots_preview_bin_trafaret = t.Dict(
    {
        t.Key("start_date"): parse_time,
        t.Key("end_date"): parse_time,
        t.Key("actual"): t.Or(t.Float, t.Null),
        t.Key("predicted"): t.Or(t.Float, t.Null),
    }
).ignore_extra("*")

_anomaly_over_time_plots_preview_bin_trafaret = t.Dict(
    {t.Key("start_date"): parse_time, t.Key("end_date"): parse_time}
).ignore_extra("*")

_calendar_event_trafaret = t.Dict(
    {t.Key("name"): String, t.Key("date"): parse_time, t.Key("series_id"): t.Or(String, t.Null)}
).ignore_extra("*")

_statistics_trafaret = t.Dict({t.Key("durbin_watson"): t.Or(t.Float, t.Null)}).ignore_extra("*")


class DatetimeTrendPlotsAPIObject(APIObject):  # pylint: disable=missing-class-docstring
    @classmethod
    def from_server_data(cls, data, *args, **kwargs):
        """
        Instantiate an object of this class using the data directly from the server,
        meaning that the keys may have the wrong camel casing.

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place
        """
        return cls.from_data(from_api(data, keep_null_keys=True))


class DatetimeTrendPlotsMetadataObject(DatetimeTrendPlotsAPIObject):
    def _get_status(self, backtest=0, source=SOURCE_TYPE.VALIDATION):
        try:
            if backtest == DATA_SUBSET.HOLDOUT:
                return self.holdout_statuses[source]
            else:
                return self.backtest_statuses[backtest][source]
        except (TypeError, KeyError, IndexError):
            return None


class AccuracyOverTimePlotsMetadata(DatetimeTrendPlotsMetadataObject):
    """
    Accuracy over Time metadata for datetime model.

    .. versionadded:: v2.25

    Attributes
    ----------
    project_id: string
        The project ID.
    model_id: string
        The model ID.
    forecast_distance: int or None
        The forecast distance for which the metadata was retrieved. None for OTV projects.
    resolutions: list of string
        A list of ``datarobot.enums.DATETIME_TREND_PLOTS_RESOLUTION``, which represents
        available time resolutions for which plots can be retrieved.
    backtest_metadata: list of dict
        List of backtest metadata dicts.
        The list index of metadata dict is the backtest index.
        See backtest/holdout metadata info in `Notes` for more details.
    holdout_metadata: dict
        Holdout metadata dict. See backtest/holdout metadata info in `Notes` for more details.
    backtest_statuses: list of dict
        List of backtest statuses dict. The list index of status dict is the backtest index.
        See backtest/holdout status info in `Notes` for more details.
    holdout_statuses: dict
        Holdout status dict. See backtest/holdout status info in `Notes` for more details.

    Notes
    -----

    Backtest/holdout status is a dict containing the following:

    * training: string
        Status backtest/holdout training. One of ``datarobot.enums.DATETIME_TREND_PLOTS_STATUS``
    * validation: string
        Status backtest/holdout validation. One of ``datarobot.enums.DATETIME_TREND_PLOTS_STATUS``

    Backtest/holdout metadata is a dict containing the following:

    * training: dict
        Start and end dates for the backtest/holdout training.
    * validation: dict
        Start and end dates for the backtest/holdout validation.

    Each dict in the `training` and `validation` in backtest/holdout metadata is structured like:

    * start_date: datetime.datetime or None
        The datetime of the start of the chart data (inclusive). None if chart data is not computed.
    * end_date: datetime.datetime or None
        The datetime of the end of the chart data (exclusive). None if chart data is not computed.
    """

    _converter = t.Dict(
        {
            t.Key("project_id"): String,
            t.Key("model_id"): String,
            t.Key("forecast_distance"): t.Or(Int, t.Null),
            t.Key("resolutions"): t.List(t.Enum(*DATETIME_TREND_PLOTS_RESOLUTION.ALL)),
            t.Key("backtest_metadata"): t.List(_backtest_holdout_metadata_trafaret),
            t.Key("holdout_metadata"): _backtest_holdout_metadata_trafaret,
            t.Key("backtest_statuses"): t.List(_accuracy_anomaly_over_time_plots_status_trafaret),
            t.Key("holdout_statuses"): _accuracy_anomaly_over_time_plots_status_trafaret,
        }
    ).ignore_extra("*")

    def __init__(
        self,
        project_id,
        model_id,
        forecast_distance,
        resolutions,
        backtest_metadata,
        holdout_metadata,
        backtest_statuses,
        holdout_statuses,
    ):
        self.project_id = project_id
        self.model_id = model_id
        self.forecast_distance = forecast_distance
        self.resolutions = resolutions
        self.backtest_metadata = backtest_metadata
        self.holdout_metadata = holdout_metadata
        self.backtest_statuses = backtest_statuses
        self.holdout_statuses = holdout_statuses

    def __repr__(self):
        return "{}(project_id={}, model_id={}, forecast_distance={})".format(
            self.__class__.__name__,
            self.project_id,
            self.model_id,
            self.forecast_distance,
        )


class AccuracyOverTimePlot(DatetimeTrendPlotsAPIObject):
    """
    Accuracy over Time plot for datetime model.

    .. versionadded:: v2.25

    Attributes
    ----------
    project_id: string
        The project ID.
    model_id: string
        The model ID.
    resolution: string
        The resolution that is used for binning.
        One of ``datarobot.enums.DATETIME_TREND_PLOTS_RESOLUTION``
    start_date: datetime.datetime
        The datetime of the start of the chart data (inclusive).
    end_date: datetime.datetime
        The datetime of the end of the chart data (exclusive).
    bins: list of dict
        List of plot bins. See bin info in `Notes` for more details.
    statistics: dict
        Statistics for plot. See statistics info in `Notes` for more details.
    calendar_events: list of dict
        List of calendar events for the plot. See calendar events info in `Notes` for more details.

    Notes
    -----

    Bin is a dict containing the following:

    * start_date: datetime.datetime
        The datetime of the start of the bin (inclusive).
    * end_date: datetime.datetime
        The datetime of the end of the bin (exclusive).
    * actual: float or None
        Average actual value of the target in the bin. None if there are no entries in the bin.
    * predicted: float or None
        Average prediction of the model in the bin. None if there are no entries in the bin.
    * frequency: int or None
        Indicates number of values averaged in bin.

    Statistics is a dict containing the following:

    * durbin_watson: float or None
        The Durbin-Watson statistic for the chart data.
        Value is between 0 and 4. Durbin-Watson statistic
        is a test statistic used to detect the presence of
        autocorrelation at lag 1 in the residuals (prediction errors)
        from a regression analysis. More info
        https://wikipedia.org/wiki/Durbin%E2%80%93Watson_statistic

    Calendar event is a dict containing the following:

    * name: string
        Name of the calendar event.
    * date: datetime
        Date of the calendar event.
    * series_id: string or None
        The series ID for the event. If this event does not specify a series ID,
        then this will be None, indicating that the event applies to all series.
    """

    _converter = t.Dict(
        {
            t.Key("project_id"): String,
            t.Key("model_id"): String,
            t.Key("start_date"): parse_time,
            t.Key("end_date"): parse_time,
            t.Key("resolution"): t.Enum(*DATETIME_TREND_PLOTS_RESOLUTION.ALL),
            t.Key("bins"): t.List(_accuracy_over_time_plots_bin_trafaret),
            t.Key("statistics"): _statistics_trafaret,
            t.Key("calendar_events"): t.List(_calendar_event_trafaret),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        project_id,
        model_id,
        start_date,
        end_date,
        resolution,
        bins,
        statistics,
        calendar_events,
    ):
        self.project_id = project_id
        self.model_id = model_id
        self.start_date = start_date
        self.end_date = end_date
        self.resolution = resolution
        self.bins = bins
        self.statistics = statistics
        self.calendar_events = calendar_events

    def __repr__(self):
        return "{}(project_id={}, model_id={}, start_date={}, end_date={})".format(
            self.__class__.__name__,
            self.project_id,
            self.model_id,
            datetime_to_string(self.start_date, ensure_rfc_3339=True),
            datetime_to_string(self.end_date, ensure_rfc_3339=True),
        )


class AccuracyOverTimePlotPreview(DatetimeTrendPlotsAPIObject):
    """
    Accuracy over Time plot preview for datetime model.

    .. versionadded:: v2.25

    Attributes
    ----------
    project_id: string
        The project ID.
    model_id: string
        The model ID.
    start_date: datetime.datetime
        The datetime of the start of the chart data (inclusive).
    end_date: datetime.datetime
        The datetime of the end of the chart data (exclusive).
    bins: list of dict
        List of plot bins. See bin info in `Notes` for more details.

    Notes
    -----

    Bin is a dict containing the following:

    * start_date: datetime.datetime
        The datetime of the start of the bin (inclusive).
    * end_date: datetime.datetime
        The datetime of the end of the bin (exclusive).
    * actual: float or None
        Average actual value of the target in the bin. None if there are no entries in the bin.
    * predicted: float or None
        Average prediction of the model in the bin. None if there are no entries in the bin.

    """

    _converter = t.Dict(
        {
            t.Key("project_id"): String,
            t.Key("model_id"): String,
            t.Key("start_date"): parse_time,
            t.Key("end_date"): parse_time,
            t.Key("bins"): t.List(_datetime_trend_plots_preview_bin_trafaret),
        }
    ).ignore_extra("*")

    def __init__(self, project_id, model_id, start_date, end_date, bins):
        self.project_id = project_id
        self.model_id = model_id
        self.start_date = start_date
        self.end_date = end_date
        self.bins = bins

    def __repr__(self):
        return "{}(project_id={}, model_id={})".format(
            self.__class__.__name__, self.project_id, self.model_id
        )


class ForecastVsActualPlotsMetadata(DatetimeTrendPlotsMetadataObject):
    """
    Forecast vs Actual plots metadata for datetime model.

    .. versionadded:: v2.25

    Attributes
    ----------
    project_id: string
        The project ID.
    model_id: string
        The model ID.
    resolutions: list of string
        A list of ``datarobot.enums.DATETIME_TREND_PLOTS_RESOLUTION``, which represents
        available time resolutions for which plots can be retrieved.
    backtest_metadata: list of dict
        List of backtest metadata dicts.
        The list index of metadata dict is the backtest index.
        See backtest/holdout metadata info in `Notes` for more details.
    holdout_metadata: dict
        Holdout metadata dict. See backtest/holdout metadata info in `Notes` for more details.
    backtest_statuses: list of dict
        List of backtest statuses dict. The list index of status dict is the backtest index.
        See backtest/holdout status info in `Notes` for more details.
    holdout_statuses: dict
        Holdout status dict. See backtest/holdout status info in `Notes` for more details.

    Notes
    -----

    Backtest/holdout status is a dict containing the following:

    * training: dict
        Dict containing each of ``datarobot.enums.DATETIME_TREND_PLOTS_STATUS`` as dict key,
        and list of forecast distances for particular status as dict value.

    * validation: dict
        Dict containing each of ``datarobot.enums.DATETIME_TREND_PLOTS_STATUS`` as dict key,
        and list of forecast distances for particular status as dict value.

    Backtest/holdout metadata is a dict containing the following:

    * training: dict
        Start and end dates for the backtest/holdout training.
    * validation: dict
        Start and end dates for the backtest/holdout validation.

    Each dict in the `training` and `validation` in backtest/holdout metadata is structured like:

    * start_date: datetime.datetime or None
        The datetime of the start of the chart data (inclusive). None if chart data is not computed.
    * end_date: datetime.datetime or None
        The datetime of the end of the chart data (exclusive). None if chart data is not computed.
    """

    _converter = t.Dict(
        {
            t.Key("project_id"): String,
            t.Key("model_id"): String,
            t.Key("resolutions"): t.List(t.Enum(*DATETIME_TREND_PLOTS_RESOLUTION.ALL)),
            t.Key("backtest_metadata"): t.List(_backtest_holdout_metadata_trafaret),
            t.Key("holdout_metadata"): _backtest_holdout_metadata_trafaret,
            t.Key("backtest_statuses"): t.List(_forecast_vs_actual_status_trafaret),
            t.Key("holdout_statuses"): _forecast_vs_actual_status_trafaret,
        }
    ).ignore_extra("*")

    def __init__(
        self,
        project_id,
        model_id,
        resolutions,
        backtest_metadata,
        holdout_metadata,
        backtest_statuses,
        holdout_statuses,
    ):
        self.project_id = project_id
        self.model_id = model_id
        self.resolutions = resolutions
        self.backtest_metadata = backtest_metadata
        self.holdout_metadata = holdout_metadata
        self.backtest_statuses = backtest_statuses
        self.holdout_statuses = holdout_statuses

    def __repr__(self):
        return "{}(project_id={}, model_id={})".format(
            self.__class__.__name__,
            self.project_id,
            self.model_id,
        )


class ForecastVsActualPlot(DatetimeTrendPlotsAPIObject):
    """
    Forecast vs Actual plot for datetime model.

    .. versionadded:: v2.25

    Attributes
    ----------
    project_id: string
        The project ID.
    model_id: string
        The model ID.
    forecast_distances: list of int
        A list of forecast distances that were retrieved.
    resolution: string
        The resolution that is used for binning.
        One of ``datarobot.enums.DATETIME_TREND_PLOTS_RESOLUTION``
    start_date: datetime.datetime
        The datetime of the start of the chart data (inclusive).
    end_date: datetime.datetime
        The datetime of the end of the chart data (exclusive).
    bins: list of dict
        List of plot bins. See bin info in `Notes` for more details.
    calendar_events: list of dict
        List of calendar events for the plot. See calendar events info in `Notes` for more details.

    Notes
    -----

    Bin is a dict containing the following:

    * start_date: datetime.datetime
        The datetime of the start of the bin (inclusive).
    * end_date: datetime.datetime
        The datetime of the end of the bin (exclusive).
    * actual: float or None
        Average actual value of the target in the bin. None if there are no entries in the bin.
    * forecasts: list of float
        A list of average forecasts for the model for each forecast distance.
        Empty if there are no forecasts in the bin.
        Each index in the `forecasts` list maps to `forecastDistances` list index.
    * error: float or None
        Average absolute residual value of the bin.
        None if there are no entries in the bin.
    * normalized_error: float or None
        Normalized average absolute residual value of the bin.
        None if there are no entries in the bin.
    * frequency: int or None
        Indicates number of values averaged in bin.

    Calendar event is a dict containing the following:

    * name: string
        Name of the calendar event.
    * date: datetime
        Date of the calendar event.
    * series_id: string or None
        The series ID for the event. If this event does not specify a series ID,
        then this will be None, indicating that the event applies to all series.
    """

    _converter = t.Dict(
        {
            t.Key("project_id"): String,
            t.Key("model_id"): String,
            t.Key("forecast_distances"): t.List(Int),
            t.Key("start_date"): parse_time,
            t.Key("end_date"): parse_time,
            t.Key("resolution"): t.Enum(*DATETIME_TREND_PLOTS_RESOLUTION.ALL),
            t.Key("bins"): t.List(_forecast_vs_actual_plots_bin_trafaret),
            t.Key("calendar_events"): t.List(_calendar_event_trafaret),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        project_id,
        model_id,
        forecast_distances,
        start_date,
        end_date,
        resolution,
        bins,
        calendar_events,
    ):
        self.project_id = project_id
        self.model_id = model_id
        self.forecast_distances = forecast_distances
        self.start_date = start_date
        self.end_date = end_date
        self.resolution = resolution
        self.bins = bins
        self.calendar_events = calendar_events

    def __repr__(self):
        return "{}(project_id={}, model_id={}, start_date={}, end_date={})".format(
            self.__class__.__name__,
            self.project_id,
            self.model_id,
            datetime_to_string(self.start_date, ensure_rfc_3339=True),
            datetime_to_string(self.end_date, ensure_rfc_3339=True),
        )


class ForecastVsActualPlotPreview(DatetimeTrendPlotsAPIObject):
    """
    Forecast vs Actual plot preview for datetime model.

    .. versionadded:: v2.25

    Attributes
    ----------
    project_id: string
        The project ID.
    model_id: string
        The model ID.
    start_date: datetime.datetime
        The datetime of the start of the chart data (inclusive).
    end_date: datetime.datetime
        The datetime of the end of the chart data (exclusive).
    bins: list of dict
        List of plot bins. See bin info in `Notes` for more details.

    Notes
    -----

    Bin is a dict containing the following:

    * start_date: datetime.datetime
        The datetime of the start of the bin (inclusive).
    * end_date: datetime.datetime
        The datetime of the end of the bin (exclusive).
    * actual: float or None
        Average actual value of the target in the bin. None if there are no entries in the bin.
    * predicted: float or None
        Average prediction of the model in the bin. None if there are no entries in the bin.

    """

    _converter = t.Dict(
        {
            t.Key("project_id"): String,
            t.Key("model_id"): String,
            t.Key("start_date"): parse_time,
            t.Key("end_date"): parse_time,
            t.Key("bins"): t.List(_datetime_trend_plots_preview_bin_trafaret),
        }
    ).ignore_extra("*")

    def __init__(self, project_id, model_id, start_date, end_date, bins):
        self.project_id = project_id
        self.model_id = model_id
        self.start_date = start_date
        self.end_date = end_date
        self.bins = bins

    def __repr__(self):
        return "{}(project_id={}, model_id={})".format(
            self.__class__.__name__, self.project_id, self.model_id
        )


class AnomalyOverTimePlotsMetadata(DatetimeTrendPlotsMetadataObject):
    """
    Anomaly over Time metadata for datetime model.

    .. versionadded:: v2.25

    Attributes
    ----------
    project_id: string
        The project ID.
    model_id: string
        The model ID.
    resolutions: list of string
        A list of ``datarobot.enums.DATETIME_TREND_PLOTS_RESOLUTION``, which represents
        available time resolutions for which plots can be retrieved.
    backtest_metadata: list of dict
        List of backtest metadata dicts.
        The list index of metadata dict is the backtest index.
        See backtest/holdout metadata info in `Notes` for more details.
    holdout_metadata: dict
        Holdout metadata dict. See backtest/holdout metadata info in `Notes` for more details.
    backtest_statuses: list of dict
        List of backtest statuses dict. The list index of status dict is the backtest index.
        See backtest/holdout status info in `Notes` for more details.
    holdout_statuses: dict
        Holdout status dict. See backtest/holdout status info in `Notes` for more details.

    Notes
    -----

    Backtest/holdout status is a dict containing the following:

    * training: string
        Status backtest/holdout training. One of ``datarobot.enums.DATETIME_TREND_PLOTS_STATUS``
    * validation: string
        Status backtest/holdout validation. One of ``datarobot.enums.DATETIME_TREND_PLOTS_STATUS``

    Backtest/holdout metadata is a dict containing the following:

    * training: dict
        Start and end dates for the backtest/holdout training.
    * validation: dict
        Start and end dates for the backtest/holdout validation.

    Each dict in the `training` and `validation` in backtest/holdout metadata is structured like:

    * start_date: datetime.datetime or None
        The datetime of the start of the chart data (inclusive). None if chart data is not computed.
    * end_date: datetime.datetime or None
        The datetime of the end of the chart data (exclusive). None if chart data is not computed.
    """

    _converter = t.Dict(
        {
            t.Key("project_id"): String,
            t.Key("model_id"): String,
            t.Key("resolutions"): t.List(t.Enum(*DATETIME_TREND_PLOTS_RESOLUTION.ALL)),
            t.Key("backtest_metadata"): t.List(_backtest_holdout_metadata_trafaret),
            t.Key("holdout_metadata"): _backtest_holdout_metadata_trafaret,
            t.Key("backtest_statuses"): t.List(_accuracy_anomaly_over_time_plots_status_trafaret),
            t.Key("holdout_statuses"): _accuracy_anomaly_over_time_plots_status_trafaret,
        }
    ).ignore_extra("*")

    def __init__(
        self,
        project_id,
        model_id,
        resolutions,
        backtest_metadata,
        holdout_metadata,
        backtest_statuses,
        holdout_statuses,
    ):
        self.project_id = project_id
        self.model_id = model_id
        self.resolutions = resolutions
        self.backtest_metadata = backtest_metadata
        self.holdout_metadata = holdout_metadata
        self.backtest_statuses = backtest_statuses
        self.holdout_statuses = holdout_statuses

    def __repr__(self):
        return "{}(project_id={}, model_id={})".format(
            self.__class__.__name__, self.project_id, self.model_id
        )


class AnomalyOverTimePlot(DatetimeTrendPlotsAPIObject):
    """
    Anomaly over Time plot for datetime model.

    .. versionadded:: v2.25

    Attributes
    ----------
    project_id: string
        The project ID.
    model_id: string
        The model ID.
    resolution: string
        The resolution that is used for binning.
        One of ``datarobot.enums.DATETIME_TREND_PLOTS_RESOLUTION``
    start_date: datetime.datetime
        The datetime of the start of the chart data (inclusive).
    end_date: datetime.datetime
        The datetime of the end of the chart data (exclusive).
    bins: list of dict
        List of plot bins. See bin info in `Notes` for more details.
    calendar_events: list of dict
        List of calendar events for the plot. See calendar events info in `Notes` for more details.

    Notes
    -----

    Bin is a dict containing the following:

    * start_date: datetime.datetime
        The datetime of the start of the bin (inclusive).
    * end_date: datetime.datetime
        The datetime of the end of the bin (exclusive).
    * predicted: float or None
        Average prediction of the model in the bin. None if there are no entries in the bin.
    * frequency: int or None
        Indicates number of values averaged in bin.

    Calendar event is a dict containing the following:

    * name: string
        Name of the calendar event.
    * date: datetime
        Date of the calendar event.
    * series_id: string or None
        The series ID for the event. If this event does not specify a series ID,
        then this will be None, indicating that the event applies to all series.
    """

    _converter = t.Dict(
        {
            t.Key("project_id"): String,
            t.Key("model_id"): String,
            t.Key("start_date"): parse_time,
            t.Key("end_date"): parse_time,
            t.Key("resolution"): t.Enum(*DATETIME_TREND_PLOTS_RESOLUTION.ALL),
            t.Key("bins"): t.List(_anomaly_over_time_plots_bin_trafaret),
            t.Key("calendar_events"): t.List(_calendar_event_trafaret),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        project_id,
        model_id,
        start_date,
        end_date,
        resolution,
        bins,
        calendar_events,
    ):
        self.project_id = project_id
        self.model_id = model_id
        self.start_date = start_date
        self.end_date = end_date
        self.resolution = resolution
        self.bins = bins
        self.calendar_events = calendar_events

    def __repr__(self):
        return "{}(project_id={}, model_id={}, start_date={}, end_date={})".format(
            self.__class__.__name__,
            self.project_id,
            self.model_id,
            datetime_to_string(self.start_date, ensure_rfc_3339=True),
            datetime_to_string(self.end_date, ensure_rfc_3339=True),
        )


class AnomalyOverTimePlotPreview(DatetimeTrendPlotsAPIObject):
    """
    Anomaly over Time plot preview for datetime model.

    .. versionadded:: v2.25

    Attributes
    ----------
    project_id: string
        The project ID.
    model_id: string
        The model ID.
    prediction_threshold: float
        Only bins with predictions exceeding
        this threshold are returned in the response.
    start_date: datetime.datetime
        The datetime of the start of the chart data (inclusive).
    end_date: datetime.datetime
        The datetime of the end of the chart data (exclusive).
    bins: list of dict
        List of plot bins. See bin info in `Notes` for more details.

    Notes
    -----

    Bin is a dict containing the following:

    * start_date: datetime.datetime
        The datetime of the start of the bin (inclusive).
    * end_date: datetime.datetime
        The datetime of the end of the bin (exclusive).

    """

    _converter = t.Dict(
        {
            t.Key("project_id"): String,
            t.Key("model_id"): String,
            t.Key("prediction_threshold"): t.Float,
            t.Key("start_date"): parse_time,
            t.Key("end_date"): parse_time,
            t.Key("bins"): t.List(_anomaly_over_time_plots_preview_bin_trafaret),
        }
    ).ignore_extra("*")

    def __init__(self, project_id, model_id, prediction_threshold, start_date, end_date, bins):
        self.project_id = project_id
        self.model_id = model_id
        self.prediction_threshold = prediction_threshold
        self.start_date = start_date
        self.end_date = end_date
        self.bins = bins

    def __repr__(self):
        return "{}(project_id={}, model_id={})".format(
            self.__class__.__name__, self.project_id, self.model_id
        )



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import logging
import sys


def get_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.setLevel(logging.INFO)
    logger.propagate = False
    return logger



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.


class EligibilityResult:
    """Represents whether a particular operation is supported

    For instance, a function to check whether a set of models can be blended can return an
    EligibilityResult specifying whether or not blending is supported and why it may not be
    supported.

    Attributes
    ----------
    supported : bool
        whether the operation this result represents is supported
    reason : str
        why the operation is or is not supported
    context : str
        what operation isn't supported
    """

    def __init__(self, supported: bool, reason: str = "", context: str = "") -> None:
        self.supported = supported
        self.reason = reason
        self.context = context

    def __repr__(self) -> str:
        return "{}(supported={}, reason='{}', context='{}')".format(
            self.__class__.__name__, self.supported, self.reason, self.context
        )



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot.errors import ClientError
from datarobot.models.api_object import APIObject
from datarobot.models.deployment import Deployment
from datarobot.models.genai.playground import Playground
from datarobot.models.model import Model
from datarobot.models.use_cases.use_case import UseCase
from datarobot.models.use_cases.utils import get_use_case_id, resolve_use_cases, UseCaseLike
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


def get_entity_id(entity: Union[CustomModelValidation, Deployment, Model, Playground, str]) -> str:
    """
    Get the entity ID from the entity parameter.

    Parameters
    ----------
    entity : APIObject or str
        Specifies either the entity ID or the entity.

    Returns
    -------
    id : str
        The entity ID.
    """
    return entity if isinstance(entity, str) else entity.id


custom_model_validation_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String,
        t.Key("prompt_column_name"): t.String,
        t.Key("target_column_name"): t.String,
        t.Key("deployment_id"): t.String,
        t.Key("validation_status"): t.String,
        t.Key("model_id"): t.String,
        t.Key("deployment_access_data", optional=True, default=None): t.Or(
            t.Null,
            t.Dict(
                {
                    t.Key("prediction_api_url"): t.String,
                    t.Key("datarobot_key", optional=True, default=None): t.Or(t.Null, t.String),
                    t.Key("authorization_header"): t.String,
                    t.Key("input_type"): t.String,
                    t.Key("model_type"): t.String,
                }
            ).ignore_extra("*"),
        ),
        t.Key("tenant_id"): t.String,
        t.Key("user_id"): t.String,
        t.Key("creation_date"): t.String,
        t.Key("error_message", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("deployment_name", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("user_name", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("use_case_id", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("prediction_timeout"): t.Int,
    }
).ignore_extra("*")


class CustomModelValidation(APIObject):
    """
    Validation record checking the ability of the deployment to serve
    as a custom model LLM or vector database.

    Attributes
    ----------
    prompt_column_name : str
        The column name that the deployed model expects as the input.
    target_column_name : str
        The target name that the deployed model will output.
    deployment_id : str
        ID of the deployment.
    model_id : str
        ID of the underlying deployment model.
        Can be found from the API as Deployment.model["id"].
    validation_status : str
        Can be TESTING, FAILED, or PASSED. Only PASSED is allowed for use.
    deployment_access_data : dict, optional
        Data that will be used for accessing deployment prediction server.
        Only available for deployments that pass validation. Dict fields:
        - prediction_api_url - URL for deployment prediction server.
        - datarobot_key - First of two auth headers for the prediction server.
        - authorization_header - Second of two auth headers for the prediction server.
        - input_type - Either JSON or CSV - input type model expects.
        - model_type - Target type of deployed custom model.
    tenant_id : str
        Creating user's tenant ID.
    error_message : Optional[str]
        Additional information for errored validation.
    deployment_name : Optional[str]
        The name of the deployment that is validated.
    user_name : Optional[str]
        The name of the user
    use_case_id : Optional[str]
        The ID of the Use Case associated with the validation.
    prediction_timeout: int
        The timeout, in seconds, for the prediction API used in this custom model validation.
    """

    _path: str
    _converter = custom_model_validation_trafaret

    def __init__(
        self,
        id: str,
        prompt_column_name: str,
        target_column_name: str,
        deployment_id: str,
        model_id: str,
        validation_status: str,
        deployment_access_data: Optional[Dict[str, Any]],
        tenant_id: str,
        name: str,
        creation_date: str,
        user_id: str,
        error_message: Optional[str],
        deployment_name: Optional[str],
        user_name: Optional[str],
        use_case_id: Optional[str],
        prediction_timeout: int,
    ):
        self.id = id
        self.prompt_column_name = prompt_column_name
        self.target_column_name = target_column_name
        self.deployment_id = deployment_id
        self.model_id = model_id
        self.validation_status = validation_status
        self.deployment_access_data = deployment_access_data
        self.tenant_id = tenant_id
        self.error_message = error_message
        self.name = name
        self.creation_date = creation_date
        self.user_id = user_id
        self.deployment_name = deployment_name
        self.user_name = user_name
        self.use_case_id = use_case_id
        self.prediction_timeout = prediction_timeout

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id})"

    @classmethod
    def get(cls, validation_id: str) -> CustomModelValidation:
        """
        Get the validation record by id.

        Parameters
        ----------
        validation_id : Union[CustomModelValidation, str]
            The CustomModelValidation to retrieve, either `CustomModelValidation` or validation ID.

        Returns
        -------
        CustomModelValidation
        """
        url = f"{cls._client.domain}/{cls._path}/{get_entity_id(validation_id)}/"
        response = cls._client.get(url)
        return cls.from_server_data(response.json())

    @classmethod
    def get_by_values(
        cls, prompt_column_name: str, target_column_name: str, deployment_id: str, model_id: str
    ) -> CustomModelValidation:
        """
        Get the validation record by field values.

        Parameters
        ----------
        prompt_column_name : str
            The column name the deployed model expect as the input.
        target_column_name : str
            The target name deployed model will output.
        deployment_id : str
            ID of the deployment.
        model_id : str
            ID of the underlying deployment model.

        Returns
        -------
        CustomModelValidation
        """

        url = f"{cls._client.domain}/{cls._path}/"
        params = {
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "deployment_id": deployment_id,
            "model_id": model_id,
            "order_by": "-creationDate",
            "limit": 1,
        }
        response_body = cls._client.get(url, params=params).json()
        data = response_body.get("data")

        if data is None:
            return cls.from_server_data(response_body)
        else:
            if len(data) == 0:
                raise ClientError(
                    exc_message="Custom model LLM validation not found", status_code=404
                )
            else:
                return cls.from_server_data(data[0])

    @classmethod
    def list(
        cls,
        prompt_column_name: Optional[str] = None,
        target_column_name: Optional[str] = None,
        deployment: Optional[Union[Deployment, str]] = None,
        model: Optional[Union[Model, str]] = None,
        use_cases: Optional[UseCaseLike] = None,
        playground: Optional[Union[Playground, str]] = None,
        completed_only: bool = False,
        search: Optional[str] = None,
        sort: Optional[str] = None,
    ) -> List[CustomModelValidation]:
        """
        List the validation records by field values.

        Parameters
        ----------
        prompt_column_name : Optional[str], optional
            The column name the deployed model expects as the input.
        target_column_name : Optional[str], optional
            The target name that the deployed model will output.
        deployment : Optional[Union[Deployment, str]], optional
            The returned validations are filtered to those associated with a specific deployment
            if specified, either `Deployment` or deployment ID.
        model_id : Optional[Union[Model, str]], optional
            The returned validations are filtered to those associated with a specific model
            if specified, either `Model` or model ID.
        use_cases : Optional[list[Union[UseCase, str]]], optional
            The returned validations are filtered to those associated with specific Use Cases
            if specified, either `UseCase` or Use Case IDs.
        playground_id : Optional[Union[Playground, str]], optional
            The returned validations are filtered to those used in a specific playground
            if specified, either `Playground` or playground ID.
        completed_only : bool, optional
            Whether to retrieve only completed validations.
        search : Optional[str], optional
            String for filtering validations.
            Validations that contain the string in name will be returned.
        sort : Optional[str], optional
            Property to sort validations by.
            Prefix the attribute name with a dash to sort in descending order,
            e.g. sort='-name'.
            Currently supported options are listed in ListCustomModelValidationsSortQueryParams
            but the values can differ with different platform versions.
            By default, the sort parameter is None which will result in
            validations being returned in order of creation time descending.

        Returns
        -------
        List[CustomModelValidation]
        """

        url = f"{cls._client.domain}/{cls._path}/"
        params = {
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "deployment_id": get_entity_id(deployment) if deployment else None,
            "model_id": get_entity_id(model) if model else None,
            "playground_id": get_entity_id(playground) if playground else None,
            "completed_only": completed_only,
            "search": search,
            "sort": sort if sort else "-creationDate",
        }
        params = resolve_use_cases(use_cases=use_cases, params=params)
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    @classmethod
    def create(
        cls,
        prompt_column_name: str,
        target_column_name: str,
        deployment_id: Union[Deployment, str],
        model: Optional[Union[Model, str]] = None,
        use_case: Optional[Union[UseCase, str]] = None,
        name: Optional[str] = None,
        wait_for_completion: bool = False,
        prediction_timeout: Optional[int] = None,
    ) -> CustomModelValidation:
        """
        Start the validation of deployment to serve as a vector database or LLM.

        Parameters
        ----------
        prompt_column_name : str
            The column name the deployed model expect as the input.
        target_column_name : str
            The target name that the deployed model will output.
        deployment_id : Union[Deployment, str]
            The deployment to validate, either `Deployment` or deployment ID.
        model : Optional[Union[Model, str]], optional
            The specific model within the deployment, either `Model` or model ID.
            If not specified, the underlying model ID will be derived from the
            deployment info automatically.
        use_case : Optional[Union[UseCase, str]], optional
            The Use Case to link the validation to, either `UseCase` or Use Case ID.
        name : Optional[str], optional
            The name of the validation.
        wait_for_completion : bool
            If set to True code will wait for the validation job to complete before
            returning the result (up to 10 minutes, raising timeout error after that).
            Otherwise, you can check current validation status by using
            CustomModelValidation.get with returned ID.
        prediction_timeout : Optional[int], optional
            The timeout, in seconds, for the prediction API used in this custom model validation.

        Returns
        -------
        CustomModelValidation
        """

        payload = {
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "deployment_id": get_entity_id(deployment_id),
            "model_id": get_entity_id(model) if model else None,
            "use_case_id": get_use_case_id(use_case, is_required=False),
            "name": name,
        }
        if prediction_timeout is not None:
            payload["prediction_timeout"] = prediction_timeout  # type: ignore[assignment]
        url = f"{cls._client.domain}/{cls._path}/"
        response = cls._client.post(url, data=payload)
        if wait_for_completion:
            location = wait_for_async_resolution(cls._client, response.headers["Location"])
            return cls.from_location(location)
        return cls.from_server_data(response.json())

    @classmethod
    def revalidate(cls, validation_id: str) -> CustomModelValidation:
        """
        Revalidate an unlinked custom model vector database or LLM.
        This method is useful when a deployment used as vector database or LLM is accidentally
        replaced with another model that stopped complying with the vector database or LLM
        requirements. Replace the model back and call this method instead of creating a new
        custom model validation from scratch.
        Another use case for this is when the API token used to create a validation record
        got revoked and no longer can be used by vector database / LLM to call custom model
        deployment. Calling revalidate will update the validation record with the token
        currently in use.

        Parameters
        ----------
        validation_id : str
            The ID of the CustomModelValidation for revalidation.

        Returns
        -------
        CustomModelValidation
        """
        url = f"{cls._client.domain}/{cls._path}/{validation_id}/revalidate/"
        response = cls._client.post(url)
        return cls.from_server_data(response.json())

    def update(
        self,
        name: Optional[str] = None,
        prompt_column_name: Optional[str] = None,
        target_column_name: Optional[str] = None,
        deployment: Optional[Union[Deployment, str]] = None,
        model: Optional[Union[Model, str]] = None,
        prediction_timeout: Optional[int] = None,
    ) -> CustomModelValidation:
        """
        Update a custom model validation.

        Parameters
        ----------
        name : Optional[str], optional
            The new name of the custom model validation.
        prompt_column_name : Optional[str], optional
            The new name of the prompt column.
        target_column_name : Optional[str], optional
            The new name of the target column.
        deployment : Optional[Union[Deployment, str]], optional
            The new deployment to validate.
        model : Optional[Union[Model, str]], optional
            The new model within the deployment to validate.
        prediction_timeout : Optional[int], optional
            The new timeout, in seconds, for the prediction API used in this custom model validation.

        Returns
        -------
        CustomModelValidation
        """
        payload = {
            "name": name,
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "deployment_id": get_entity_id(deployment) if deployment else None,
            "model_id": get_entity_id(model) if model else None,
            "prediction_timeout": prediction_timeout,
        }
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        response = self._client.patch(url, data=payload)
        return self.from_server_data(response.json())

    def delete(self) -> None:
        """
        Delete the custom model validation.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import List

import trafaret as t
from typing_extensions import TypedDict

from datarobot._compat import Int, String
from datarobot.enums import CHART_DATA_SOURCE
from datarobot.models.api_object import APIObject

_PercentageFloat = t.Float(gte=0, lte=1)  # 0.0 <= x <= 1.0

_ClassPercentage = t.Dict(
    {t.Key("other_class_name"): String, t.Key("percentage"): _PercentageFloat}
).ignore_extra("*")

_ClassMetrics = t.Dict(
    {
        t.Key("class_name"): String,
        # blue bars on the axis
        t.Key("actual_count"): Int,
        t.Key("predicted_count"): Int,
        # one vs all metrics
        t.Key("f1"): _PercentageFloat,
        t.Key("recall"): _PercentageFloat,
        t.Key("precision"): _PercentageFloat,
        t.Key("confusion_matrix_one_vs_all"): t.List(t.List(Int)),
        t.Key("was_actual_percentages"): t.List(_ClassPercentage),
        t.Key("was_predicted_percentages"): t.List(_ClassPercentage),
    }
).ignore_extra("*")

ConfusionChartTrafaret = t.Dict(
    {
        # available classes
        t.Key("classes"): t.List(String),
        # NxN confusion matrix
        t.Key("confusion_matrix"): t.List(t.List(Int)),
        t.Key("class_metrics"): t.List(_ClassMetrics),
    }
).ignore_extra("*")


class ClassPercentage(TypedDict):
    other_class_name: str
    percentage: float


class ClassMetric(TypedDict):
    class_name: str
    actual_count: int
    predicted_count: int
    f1: float
    recall: float
    precision: float
    confusion_matrix_one_vs_all: List[List[int]]
    was_actual_percentages: List[ClassPercentage]
    was_predicted_percentages: List[ClassPercentage]


class ConfusionChartRawData(TypedDict):
    class_metrics: List[ClassMetric]
    confusion_matrix: List[List[int]]
    classes: List[str]


class ConfusionChart(APIObject):
    """ Confusion Chart data for model.

    Notes
    -----
    ``ClassMetrics`` is a dict containing the following:

        * ``class_name`` (string) name of the class
        * ``actual_count`` (int) number of times this class is seen in the validation data
        * ``predicted_count`` (int) number of times this class has been predicted for the \
          validation data
        * ``f1`` (float) F1 score
        * ``recall`` (float) recall score
        * ``precision`` (float) precision score
        * ``was_actual_percentages`` (list of dict) one vs all actual percentages in format \
          specified below.
            * ``other_class_name`` (string) the name of the other class
            * ``percentage`` (float) the percentage of the times this class was predicted when is \
              was actually class (from 0 to 1)
        * ``was_predicted_percentages`` (list of dict) one vs all predicted percentages in format \
          specified below.
            * ``other_class_name`` (string) the name of the other class
            * ``percentage`` (float) the percentage of the times this class was actual predicted \
              (from 0 to 1)
        * ``confusion_matrix_one_vs_all`` (list of list) 2d list representing 2x2 one vs all matrix.
            * This represents the True/False Negative/Positive rates as integer for each class. \
              The data structure looks like:
            * ``[ [ True Negative, False Positive ], [ False Negative, True Positive ] ]``


    Attributes
    ----------
    source : str
        Confusion Chart data source. Can be 'validation', 'crossValidation' or 'holdout'.
    raw_data : dict
        All of the raw data for the Confusion Chart
    confusion_matrix : list of list
        The N x N confusion matrix
    classes : list
        The names of each of the classes
    class_metrics : list of dicts
        List of dicts with schema described as ``ClassMetrics`` above.
    source_model_id : str
        ID of the model this Confusion chart represents; in some cases,
        insights from the parent of a frozen model may be used

    """

    ConfusionChartWrapper = t.Dict(
        {
            t.Key("data"): ConfusionChartTrafaret,
            t.Key("source"): String,
            t.Key("source_model_id"): String,
        }
    ).ignore_extra("*")

    _converter = ConfusionChartWrapper

    def __init__(
        self, source: CHART_DATA_SOURCE, data: ConfusionChartRawData, source_model_id: str
    ):
        self.source = source
        self.raw_data = data
        self.class_metrics = data["class_metrics"]
        self.confusion_matrix = data["confusion_matrix"]
        self.classes = data["classes"]
        self.source_model_id = source_model_id

    def __repr__(self) -> str:
        return f"ConfusionChart({self.source})"



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import trafaret as t

from datarobot._compat import Int, String

UserBlueprintTaskArgumentDefinition_ = t.Dict(
    {
        t.Key("name"): String(allow_blank=False),
        t.Key("type"): String(allow_blank=False),
        t.Key("default", optional=True): t.Or(
            t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
            t.List(t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
        ),
        t.Key("values"): t.Or(
            t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
            t.List(t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
            t.Dict().allow_extra("*"),
        ),
        t.Key("tunable", optional=True): t.Bool(),
        t.Key("recommended", optional=True): t.Or(
            t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
            t.List(t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
        ),
    }
).allow_extra("*")


UserBlueprintTaskArgument_ = t.Dict(
    {
        t.Key("key"): String(allow_blank=False),
        t.Key("argument"): t.Or(UserBlueprintTaskArgumentDefinition_),
    }
)



# flake8: noqa
# because the unused imports are on purpose

import datarobot.models.genai
import datarobot.models.registry

from .application import Application
from .automated_documentation import AutomatedDocument
from .batch_monitoring_job import BatchMonitoringJob, BatchMonitoringJobDefinition
from .batch_prediction_job import BatchPredictionJob, BatchPredictionJobDefinition
from .blueprint import Blueprint, BlueprintChart, BlueprintTaskDocument, ModelBlueprintChart
from .calendar_file import CalendarFile
from .change_request import ChangeRequest, ChangeRequestReview
from .cluster import Cluster
from .cluster_insight import ClusterInsight
from .compliance_doc_template import ComplianceDocTemplate
from .connector import Connector
from .credential import Credential
from .custom_model import CustomInferenceModel
from .custom_model_test import CustomModelTest
from .custom_model_version import (
    CustomModelVersion,
    CustomModelVersionConversion,
    CustomModelVersionDependencyBuild,
)
from .custom_task import CustomTask
from .custom_task_version import CustomTaskVersion
from .data_engine_query_generator import DataEngineQueryGenerator
from .data_slice import DataSlice, DataSliceSizeInfo
from .data_source import DataSource, DataSourceParameters
from .data_store import DataStore
from .dataset import Dataset, DatasetDetails
from .deployment import Deployment
from .driver import DataDriver
from .execution_environment import ExecutionEnvironment
from .execution_environment_version import ExecutionEnvironmentVersion
from .external_baseline_validation import ExternalBaselineValidationInfo
from .external_dataset_scores_insights import (
    ExternalConfusionChart,
    ExternalLiftChart,
    ExternalMulticlassLiftChart,
    ExternalResidualsChart,
    ExternalRocCurve,
    ExternalScores,
)
from .feature import (
    DatasetFeature,
    DatasetFeatureHistogram,
    Feature,
    FeatureHistogram,
    FeatureLineage,
    InteractionFeature,
    ModelingFeature,
    MulticategoricalHistogram,
)
from .feature_association_matrix import (
    FeatureAssociationFeaturelists,
    FeatureAssociationMatrix,
    FeatureAssociationMatrixDetails,
)
from .feature_effect import (
    FeatureEffectMetadata,
    FeatureEffectMetadataDatetime,
    FeatureEffectMetadataDatetimePerBacktest,
    FeatureEffects,
    FeatureEffectsMulticlass,
)
from .featurelist import DatasetFeaturelist, Featurelist, ModelingFeaturelist
from .job import FeatureImpactJob, Job, TrainingPredictionsJob
from .key_values import KeyValue
from .model import (
    BlenderModel,
    ClusteringModel,
    CombinedModel,
    DatetimeModel,
    FrozenModel,
    GenericModel,
    Model,
    ModelParameters,
    PrimeModel,
    RatingTableModel,
)
from .model_registry import RegisteredModel, RegisteredModelVersion
from .modeljob import ModelJob
from .pairwise_statistics import (
    PairwiseConditionalProbabilities,
    PairwiseCorrelations,
    PairwiseJointProbabilities,
)
from .payoff_matrix import PayoffMatrix
from .predict_job import PredictJob
from .prediction_dataset import PredictionDataset
from .prediction_environment import PredictionEnvironment
from .prediction_explanations import (
    ClassListMode,
    PredictionExplanations,
    PredictionExplanationsInitialization,
    TopPredictionsMode,
)
from .prediction_server import PredictionServer
from .predictions import Predictions
from .prime_file import PrimeFile
from .project import Project
from .rating_table import RatingTable
from .recommended_model import ModelRecommendation
from .relationships_configuration import RelationshipsConfiguration
from .ruleset import Ruleset
from .secondary_dataset import SecondaryDatasetConfigurations
from .segmentation import SegmentationTask, SegmentInfo
from .shap_impact import ShapImpact
from .shap_matrix import ShapMatrix
from .shap_matrix_job import ShapMatrixJob
from .sharing import SharingAccess, SharingRole
from .status_check_job import JobStatusResult, StatusCheckJob
from .training_predictions import TrainingPredictions
from .types import (
    AnomalyAssessmentDataPoint,
    AnomalyAssessmentPreviewBin,
    AnomalyAssessmentRecordMetadata,
    RegionExplanationsData,
    RocCurveEstimatedMetric,
    ShapleyFeatureContribution,
)
from .use_cases.use_case import UseCase
from .user_blueprints import UserBlueprint



# flake8: noqa
from .feature_association_featurelists import FeatureAssociationFeaturelists
from .feature_association_matrix import FeatureAssociationMatrix
from .feature_association_matrix_details import FeatureAssociationMatrixDetails



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, Dict, Iterable, List, Mapping, Optional, Union

import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject
from datarobot.utils import from_api

ResidualsTrafaret = {
    t.Key("residual_mean"): t.Float,
    t.Key("coefficient_of_determination"): t.Float,
    t.Key("standard_deviation", optional=True): t.Float,
}


class ResidualsChart(APIObject):
    """Residual analysis chart data for model.

    .. versionadded:: v2.18

    This data is calculated over a randomly downsampled subset of the source data
    (capped at 1000 rows).

    Notes
    -----

    ``ResidualsChartRow`` is a list of floats and ints containing the following:
        * Element 0 (float) is the actual target value for the source data row.
        * Element 1 (float) is the predicted target value for that row.
        * Element 2 (float) is the error rate of predicted - actual and is optional.
        * Element 3 (int) is the row number in the source dataset from which the values
          were selected and is optional.

    Attributes
    ----------
    source : str
        Lift chart data source. Can be 'validation', 'crossValidation' or 'holdout'.
    data : list
        List of lists with schema described as ``ResidualsChartRow`` above.
    coefficient_of_determination : float
        The r-squared value for the downsampled dataset
    residual_mean : float
        The arithmetic mean of the residual (predicted value minus actual value)
    source_model_id : str
        ID of the model this chart represents; in some cases,
        insights from the parent of a frozen model may be used
    standard_deviation : float
        standard_deviation of residual values
    data_slice_id: string or None
        The slice to retrieve Feature Effects for; if None, retrieve unsliced data.
    """

    _converter = (
        t.Dict(
            {
                t.Key("source"): String,
                t.Key("data"): t.List(t.List(t.Float)),
                t.Key("source_model_id"): String,
                t.Key("data_slice_id", optional=True): t.Or(String, t.Null),
            }
        )
        .merge(ResidualsTrafaret)
        .ignore_extra("*")
    )

    def __init__(
        self,
        source: str,
        data: List[Union[float, int]],
        residual_mean: float,
        coefficient_of_determination: float,
        source_model_id: str,
        standard_deviation: Optional[float] = None,
        data_slice_id: Optional[str] = None,
    ) -> None:
        self.source = source
        self.data = data
        self.source_model_id = source_model_id
        self.coefficient_of_determination = coefficient_of_determination
        self.residual_mean = residual_mean
        self.standard_deviation = standard_deviation
        self.data_slice_id = data_slice_id

    def __repr__(self) -> str:
        return f"ResidualChart({self.source})"

    @staticmethod
    def _repack_insights_response(raw_server_record: Dict[str, Any]) -> Dict[str, Any]:
        """Repack the JSON sent by the GET /insights/ endpoint to match the format expected by the
        insight APIObject class.

        Parameters
        ----------
        raw_server_record : dict

        Returns
        -------
        server_record : dict
        """
        return {
            "source": raw_server_record["source"],
            "data": raw_server_record["data"]["data"],
            "residual_mean": raw_server_record["data"]["residualMean"],
            "coefficient_of_determination": raw_server_record["data"]["coefficientOfDetermination"],
            "source_model_id": raw_server_record["entityId"],
            "standard_deviation": raw_server_record["data"]["standardDeviation"],
            "data_slice_id": raw_server_record["dataSliceId"],
        }

    @classmethod
    # type: ignore[override]
    def from_server_data(
        cls,
        data: Dict[str, Any],
        keep_attrs: Optional[Iterable[str]] = None,
        use_insights_format: bool = False,
        **kwargs: Mapping[str, Any],
    ) -> "ResidualsChart":
        """
        Overwrite APIObject.from_server_data to handle residuals chart data retrieved
        from either legacy URL or /insights/ new URL.
        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place.
        keep_attrs : iterable
            List, set or tuple of the dotted namespace notations for attributes to keep within the
            object structure even if their values are None
        use_insights_format : bool, optional
            Whether to repack the data from the format used in the GET /insights/residuals/ URL
            to the format used in the legacy URL.
        """
        if use_insights_format:
            data = cls._repack_insights_response(data)

        case_converted = from_api(data)
        return cls.from_data(case_converted)


class SlicedResidualsChart(ResidualsChart):
    """Wrapper around ResidualsChart to override `from_server_data` method"""

    @classmethod
    # type: ignore[override]
    def from_server_data(
        cls,
        data: Dict[str, Any],
        keep_attrs: Optional[Iterable[str]] = None,
        use_insights_format: bool = True,
        **kwargs: Mapping[str, Any],
    ) -> "ResidualsChart":
        """
        Overwrite ResidualsChart.from_server_data to set `use_insights_format=True` by default
        This is necessary for the correct transformation of data received from /insights endpoints

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place.
        keep_attrs : iterable
            List, set or tuple of the dotted namespace notations for attributes to keep within the
            object structure even if their values are None
        use_insights_format : bool, optional
            Whether to repack the data from the format used in the GET /insights/liftChart/ URL
            to the format used in the legacy URL.
        """
        if data.get("count"):
            # it's a list
            data = data["data"][0]

        return super().from_server_data(
            data=data, keep_attrs=keep_attrs, use_insights_format=use_insights_format, **kwargs
        )



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, cast, List

import numpy
import trafaret as t

from datarobot.insights.base import BaseInsight


class ShapMatrix(BaseInsight):
    """Class for SHAP Matrix calculations. Use the standard methods of BaseInsight to compute and
    retrieve SHAP matrices:
    * compute: submit a request to compute a SHAP matrix, and return immediately
    * create: submit a request to compute a SHAP matrix, and wait for it to finish
    * list: retrieve all ShapMatrix results for a model, possibly on multiple datasets or data slices.
    """

    INSIGHT_NAME = "shapMatrix"
    INSIGHT_DATA = {
        t.Key("index"): t.List(t.Int()),
        t.Key("link_function"): t.String(),
        t.Key("base_value"): t.Float(),
        t.Key("colnames"): t.List(t.String()),
        t.Key("matrix"): t.List(t.List(t.Or(t.Int(), t.Float()))),
    }

    @property
    def matrix(self) -> Any:  # numpy.types.NDArray is not compatible with sphinx docs.
        """SHAP matrix values."""
        return numpy.array(self.data["matrix"])

    @property
    def base_value(self) -> float:
        """SHAP base value for the matrix values"""
        return cast(float, self.data["base_value"])

    @property
    def columns(self) -> List[str]:
        """List of columns associated with the SHAP matrix"""
        return cast(List[str], self.data["colnames"])

    @property
    def link_function(self) -> str:
        """Link function used to generate the SHAP matrix"""
        return cast(str, self.data["link_function"])



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from io import StringIO
from typing import List, Optional, TYPE_CHECKING

import pandas as pd
import trafaret as t

from datarobot import errors
from datarobot._compat import String
from datarobot.enums import DEFAULT_TIMEOUT
from datarobot.models.api_object import APIObject
from datarobot.utils import deprecation, get_id_from_response
from datarobot.utils.pagination import unpaginate

if TYPE_CHECKING:
    from datarobot.models.shap_matrix_job import ShapMatrixJob


class ShapMatrix(APIObject):
    """
    Represents SHAP based prediction explanations and provides access to score values.

    Attributes
    ----------
    project_id : str
        id of the project the model belongs to
    shap_matrix_id : str
        id of the generated SHAP matrix
    model_id : str
        id of the model used to
    dataset_id : str
         id of the prediction dataset SHAP values were computed for

    Examples
    --------
    .. code-block:: python

        import datarobot as dr

        # request SHAP matrix calculation
        shap_matrix_job = dr.ShapMatrix.create(project_id, model_id, dataset_id)
        shap_matrix = shap_matrix_job.get_result_when_complete()

        # list available SHAP matrices
        shap_matrices = dr.ShapMatrix.list(project_id)
        shap_matrix = shap_matrices[0]

        # get SHAP matrix as dataframe
        shap_matrix_values = shap_matrix.get_as_dataframe()
    """

    _path = "projects/{}/shapMatrices/"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("project_id"): String(),
            t.Key("model_id"): String(),
            t.Key("dataset_id"): String(),
        }
    ).allow_extra("*")

    @deprecation.deprecated(
        deprecated_since_version="v3.4",
        will_remove_version="v3.6",
        message="This class is deprecated, please use 'datarobot.insights.ShapMatrix' instead.",
    )
    def __init__(
        self,
        project_id: str,
        id: str,
        model_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
    ) -> None:
        self.project_id = project_id
        self.model_id = model_id
        self.dataset_id = dataset_id
        self.id = id

    def __repr__(self) -> str:
        template = "{}(id={!r}, project_id={!r}, model_id={!r}, dataset_id={!r})"
        return template.format(
            type(self).__name__,
            self.id,
            self.project_id,
            self.model_id,
            self.dataset_id,
        )

    @classmethod
    @deprecation.deprecated(
        deprecated_since_version="v3.4",
        will_remove_version="v3.6",
        message="This class is deprecated, please use 'datarobot.insights.ShapMatrix.create' instead.",
    )
    def create(cls, project_id: str, model_id: str, dataset_id: str) -> ShapMatrixJob:
        """Calculate SHAP based prediction explanations against previously uploaded dataset.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        model_id : str
            id of the model for which prediction explanations are requested
        dataset_id : str
            id of the prediction dataset for which prediction explanations are requested (as
            uploaded from Project.upload_dataset)

        Returns
        -------
        job : ShapMatrixJob
            The job computing the SHAP based prediction explanations

        Raises
        ------
        ClientError
            If the server responded with 4xx status. Possible reasons are project, model or dataset
            don't exist, user is not allowed or model doesn't support SHAP based prediction
            explanations
        ServerError
            If the server responded with 5xx status
        """
        data = {"model_id": model_id, "dataset_id": dataset_id}
        url = f"projects/{project_id}/shapMatrices/"
        response = cls._client.post(url, data=data)
        job_id = get_id_from_response(response)
        from .shap_matrix_job import (  # pylint: disable=import-outside-toplevel,cyclic-import
            ShapMatrixJob,
        )

        return ShapMatrixJob.get(
            project_id=project_id, job_id=job_id, model_id=model_id, dataset_id=dataset_id
        )

    @classmethod
    def from_location(  # type: ignore[override] # pylint: disable=arguments-renamed
        cls,
        location: str,
        model_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
    ) -> ShapMatrix:
        head, tail = location.split("/shapMatrices/", 1)
        project_id, id = head.split("/")[-1], tail.split("/")[0]
        return cls(  # type: ignore[no-any-return]
            project_id=project_id, id=id, model_id=model_id, dataset_id=dataset_id
        )

    @classmethod
    @deprecation.deprecated(
        deprecated_since_version="v3.4",
        will_remove_version="v3.6",
        message="This class is deprecated, please use 'datarobot.insights.ShapMatrix.list' instead.",
    )
    def list(cls, project_id: str) -> List[ShapMatrix]:
        """
        Fetch all the computed SHAP prediction explanations for a project.

        Parameters
        ----------
        project_id : str
            id of the project

        Returns
        -------
        List of ShapMatrix
            A list of :py:class:`ShapMatrix <datarobot.models.ShapMatrix>` objects

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(
            initial_url=cls._path.format(project_id), initial_params=None, client=cls._client
        )
        result = [cls.from_server_data(item) for item in data]
        return result

    @classmethod
    @deprecation.deprecated(
        deprecated_since_version="v3.4",
        will_remove_version="v3.6",
        message="This class is deprecated, please use 'datarobot.insights.ShapMatrix.list' instead.",
    )
    def get(cls, project_id: str, id: str) -> ShapMatrix:
        """
        Retrieve the specific SHAP matrix.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        id : str
            id of the SHAP matrix

        Returns
        -------
        :py:class:`ShapMatrix <datarobot.models.ShapMatrix>` object representing specified record
        """
        return cls(project_id=project_id, id=id)  # type: ignore[no-any-return]

    @deprecation.deprecated(
        deprecated_since_version="v3.4",
        will_remove_version="v3.6",
        message="This class is deprecated, please use 'datarobot.insights.ShapMatrix' instead.",
    )
    def get_as_dataframe(self, read_timeout: int = DEFAULT_TIMEOUT.READ) -> pd.DataFrame:
        """
        Retrieve SHAP matrix values as dataframe.

        Returns
        -------
        dataframe : pandas.DataFrame
            A dataframe with SHAP scores
        read_timeout : int (optional, default 60)
            .. versionadded:: 2.29

            Wait this many seconds for the server to respond.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = self._path.format(self.project_id) + f"{self.id}/"
        resp = self._client.get(
            path, headers={"Accept": "text/csv"}, stream=True, timeout=read_timeout
        )
        if resp.status_code == 200:
            content = resp.content.decode("utf-8")
            return pd.read_csv(StringIO(content), index_col=0, encoding="utf-8")
        else:
            raise errors.ServerError(
                f"Server returned unknown status code: {resp.status_code}",
                resp.status_code,
            )



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
# pylint: disable=too-many-lines
from __future__ import annotations

import json
from typing import Any, cast, List, Optional, TYPE_CHECKING, Union

from requests import Response

from datarobot.models.api_object import APIObject
from datarobot.utils.pagination import unpaginate

from .trafarets import (
    UserBlueprint_,
    UserBlueprintAddToMenuResponse_,
    UserBlueprintCatalogSearchItem_,
    UserBlueprintSharedRolesListResponseValidator_,
    UserBlueprintsInputTypesResponse_,
    UserBlueprintsValidateTaskParametersResponse_,
    UserBlueprintTasksResponse_,
    UserBlueprintValidationResponse_,
)

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class UserBlueprintsHexColumnNameLookupEntryDict(TypedDict):
        colname: str
        hex: str
        project_id: Optional[str]


DEFAULT_BATCH_SIZE = 100


class HumanReadable:
    """
    A simple mixin to provide human readable class representation.
    """

    def __repr__(self) -> str:
        result = []

        for key, value in self.__dict__.items():
            try:
                converted_value = value
                if isinstance(value, str):
                    converted_value = value.encode("utf-8").decode("utf-8")
                result += [f"    {key}={converted_value.__repr__()},"]
            except (UnicodeEncodeError, UnicodeDecodeError):
                pass
        return f"{self.__class__.__name__}" + "(\n{}\n)".format("\n".join(result))


class CompareWithJSON:
    def as_json(self) -> Any:
        return json.loads(json.dumps(self, default=lambda i: getattr(i, "__dict__", str(i))))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, CompareWithJSON):
            return False

        return cast(bool, self.as_json() == other.as_json())


class UserBlueprint(APIObject, HumanReadable, CompareWithJSON):
    """
    A representation of a blueprint which may be modified by the user,
    saved to a user's AI Catalog, trained on projects, and shared with others.

    It is recommended to install the python library called ``datarobot_bp_workshop``,
    available via ``pip``, for the best experience when building blueprints.

    Please refer to ``http://blueprint-workshop.datarobot.com`` for tutorials,
    examples, and other documentation.

    Parameters
    ----------
    blender: bool
        Whether the blueprint is a blender.
    blueprint_id: string
        The deterministic id of the blueprint, based on its content.
    custom_task_version_metadata: list[list[string]], Optional
        An association of custom entity ids and task ids.
    diagram: string
        The diagram used by the UI to display the blueprint.
    features: list[string]
        A list of the names of tasks used in the blueprint.
    features_text: string
        A description of the blueprint via the names of tasks used.
    hex_column_name_lookup: list[UserBlueprintsHexColumnNameLookupEntry], Optional
        A lookup between hex values and data column names used in the blueprint.
    icons: list[int]
        The icon(s) associated with the blueprint.
    insights: string
        An indication of the insights generated by the blueprint.
    is_time_series: bool (Default=False)
        Whether the blueprint contains time-series tasks.
    model_type: string
        The generated or provided title of the blueprint.
    project_id: string, Optional
        The id of the project the blueprint was originally created with, if applicable.
    reference_model: bool (Default=False)
        Whether the blueprint is a reference model.
    shap_support: bool (Default=False)
        Whether the blueprint supports shapley additive explanations.
    supported_target_types: list[enum('binary', 'multiclass', 'multilabel', 'nonnegative',
    'regression', 'unsupervised', 'unsupervisedclustering')]
        The list of supported targets of the current blueprint.
    supports_gpu: bool (Default=False)
        Whether the blueprint supports execution on the GPU.
    user_blueprint_id: string
        The unique id associated with the user blueprint.
    user_id: string
        The id of the user who owns the blueprint.
    blueprint: list[dict] or list[UserBlueprintTask], Optional
        The representation of a directed acyclic graph defining a pipeline of data through tasks
        and a final estimator.
    vertex_context: list[VertexContextItem], Optional
        Info about, warnings about, and errors with a specific vertex in the blueprint.
    blueprint_context: VertexContextItemMessages
        Warnings and errors which may describe or summarize warnings or errors in the blueprint's
        vertices
    """

    _path = "userBlueprints/{userBlueprintId}/"
    _converter = UserBlueprint_

    def __init__(
        self,
        blender: bool,
        blueprint_id: str,
        diagram: str,
        features: List[str],
        features_text: str,
        icons: List[int],
        insights: str,
        model_type: str,
        supported_target_types,
        user_blueprint_id: str,
        user_id: str,
        is_time_series: bool = False,
        reference_model: bool = False,
        shap_support: bool = False,
        supports_gpu: bool = False,
        blueprint=None,
        custom_task_version_metadata=None,
        hex_column_name_lookup: Optional[List[UserBlueprintsHexColumnNameLookupEntryDict]] = None,
        project_id: Optional[str] = None,
        vertex_context: Optional[List[VertexContextItem]] = None,
        blueprint_context: Optional[VertexContextItemMessages] = None,
        **kwargs: Any,
    ) -> None:
        self.blender = blender
        self.blueprint = _init_list_of_class_or_dict(blueprint, UserBlueprintTask)
        self.blueprint_id = blueprint_id
        self.custom_task_version_metadata = custom_task_version_metadata
        self.diagram = diagram
        self.features = features
        self.features_text = features_text
        self.hex_column_name_lookup = _init_list_of_class_or_dict(
            hex_column_name_lookup, UserBlueprintHexColumnNameLookupEntry
        )
        self.icons = icons
        self.insights = insights
        self.is_time_series = is_time_series
        self.model_type = model_type
        self.project_id = project_id
        self.reference_model = reference_model
        self.shap_support = shap_support
        self.supported_target_types = supported_target_types
        self.supports_gpu = supports_gpu
        self.user_blueprint_id = user_blueprint_id
        self.user_id = user_id
        self.blueprint_context = _init_class_or_dict(blueprint_context, VertexContextItemMessages)
        self.vertex_context = _init_list_of_class_or_dict(vertex_context, VertexContextItem)

    @classmethod
    def list(
        cls, limit: int = 100, offset: int = 0, project_id: Optional[str] = None
    ) -> List[UserBlueprint]:
        """
        Fetch a list of the user blueprints the current user created

        Parameters
        ----------
        limit: int (Default=100)
            The max number of results to return.
        offset: int (Default=0)
            The number of results to skip (for pagination).
        project_id: string, Optional
            The id of the project, used to filter for original project_id.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        list[UserBlueprint]
        """
        path = "userBlueprints/"
        params = dict(
            offset=offset,
            limit=limit or DEFAULT_BATCH_SIZE,
            project_id=None if project_id is None else str(project_id),
        )
        if limit == 0:
            data = list(unpaginate(path, params, cls._client))
        else:
            data = cls._client.get(path, params=params).json()["data"]
        return [cls.from_server_data(a) for a in data]

    @classmethod
    def get(cls, user_blueprint_id: str, project_id: Optional[str] = None) -> UserBlueprint:
        """
        Retrieve a user blueprint

        Parameters
        ----------
        user_blueprint_id: string
            Used to identify a specific user-owned blueprint.
        project_id: string (optional, default is None)
            String representation of ObjectId for a given project. Used to validate selected
            columns in the user blueprint.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprint
        """
        response = cls._client.get(
            cls._path.format(userBlueprintId=user_blueprint_id),
            params=dict(edit_mode=True, decompressed_blueprint=True, project_id=project_id),
        )
        return cls.from_server_data(response.json())

    @classmethod
    def create(
        cls,
        blueprint,
        model_type: Optional[str] = None,
        project_id: Optional[str] = None,
        save_to_catalog: bool = True,
    ) -> UserBlueprint:
        """
        Create a user blueprint

        Parameters
        ----------
        blueprint: list[dict] or list[UserBlueprintTask]
            A list of tasks in the form of dictionaries which define a blueprint.
        model_type: string, Optional
            The title to give to the blueprint.
        project_id: string, Optional
            The project associated with the blueprint. Necessary in the event of project specific
            tasks, such as column selection tasks.
        save_to_catalog: bool, (Default=True)
            Whether the blueprint being created should be saved to the catalog.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprint
        """
        _path = "userBlueprints/"
        response = cls._client.post(
            _path,
            data=dict(
                blueprint=[b.as_json() if not isinstance(b, dict) else b for b in blueprint],
                project_id=None if project_id is None else str(project_id),
                model_type=model_type,
                decompressed_blueprint=True,
                save_to_catalog=save_to_catalog,
            ),
            keep_attrs=["value"],
        )
        return cls.from_server_data(response.json())

    @classmethod
    def create_from_custom_task_version_id(
        cls,
        custom_task_version_id: str,
        save_to_catalog: bool = True,
        description: Optional[str] = None,
    ) -> UserBlueprint:
        """
        Create a user blueprint with a single custom task version

        Parameters
        ----------
        custom_task_version_id: string
            Id of custom task version from which the user blueprint is created
        save_to_catalog: bool, (Default=True)
            Whether the blueprint being created should be saved to the catalog
        description: string (Default=None)
            The description for the user blueprint that will be created from the
            custom task version.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprint
        """
        _path = "userBlueprints/fromCustomTaskVersionId/"
        response = cls._client.post(
            _path,
            data=dict(
                custom_task_version_id=str(custom_task_version_id),
                save_to_catalog=save_to_catalog,
                description=description,
                decompressed_blueprint=True,
            ),
        )
        return cls.from_server_data(response.json())

    @classmethod
    def clone_project_blueprint(
        cls,
        blueprint_id: str,
        project_id: str,
        model_type: Optional[str] = None,
        save_to_catalog: bool = True,
    ) -> UserBlueprint:
        """
        Clone a blueprint from a project.

        Parameters
        ----------
        blueprint_id: string
            The id associated with the blueprint to create the user blueprint from.
        model_type: string, Optional
            The title to give to the blueprint.
        project_id: string
            The id of the project which the blueprint to copy comes from.
        save_to_catalog: bool, (Default=True)
            Whether the blueprint being created should be saved to the catalog.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprint
        """
        _path = "userBlueprints/fromBlueprintId/"
        response = cls._client.post(
            _path,
            data=dict(
                blueprint_id=blueprint_id,
                project_id=str(project_id),
                model_type=model_type,
                decompressed_blueprint=True,
                save_to_catalog=save_to_catalog,
            ),
        )
        return cls.from_server_data(response.json())

    @classmethod
    def clone_user_blueprint(
        cls,
        user_blueprint_id: str,
        model_type: Optional[str] = None,
        project_id: Optional[str] = None,
        save_to_catalog: bool = True,
    ) -> UserBlueprint:
        """
        Clone a user blueprint.

        Parameters
        ----------
        model_type: string, Optional
            The title to give to the blueprint.
        project_id: string, Optional
            String representation of ObjectId for a given project. Used to validate selected
            columns in the user blueprint.
        user_blueprint_id: string
            The id of the existing user blueprint to copy.
        save_to_catalog: bool, (Default=True)
            Whether the blueprint being created should be saved to the catalog.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprint
        """
        _path = "userBlueprints/fromUserBlueprintId/"

        response = cls._client.post(
            _path,
            data=dict(
                user_blueprint_id=str(user_blueprint_id),
                project_id=None if project_id is None else str(project_id),
                model_type=model_type,
                decompressed_blueprint=True,
                save_to_catalog=save_to_catalog,
            ),
        )
        return cls.from_server_data(response.json())

    @classmethod
    def update(
        cls,
        blueprint,
        user_blueprint_id: str,
        model_type: Optional[str] = None,
        project_id: Optional[str] = None,
        include_project_id_if_none: bool = False,
    ) -> UserBlueprint:
        """
        Update a user blueprint

        Parameters
        ----------
        blueprint: list(dict) or list(UserBlueprintTask)
            A list of tasks in the form of dictionaries which define a blueprint.
            If None, will not be passed.
        model_type: string, Optional
            The title to give to the blueprint. If None, will not be passed.
        project_id: string, Optional
            The project associated with the blueprint. Necessary in the event of project specific
            tasks, such as column selection tasks.
            If None, will not be passed. To explicitly pass None, pass True to
            `include_project_id_if_none` (useful if unlinking a blueprint from a project)
        user_blueprint_id: string
            Used to identify a specific user-owned blueprint.
        include_project_id_if_none: bool (Default=False)
            Allows project_id to be passed as None, instead of ignored.
            If set to False, will not pass `project_id` in the API request if it is set to None.
            If True, the project id will be passed even if it is set to None.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprint
        """
        kwargs = {}
        if blueprint is not None:
            kwargs["blueprint"] = [b.as_json() if not isinstance(b, dict) else b for b in blueprint]
        if model_type is not None:
            kwargs["model_type"] = model_type
        if project_id is not None or include_project_id_if_none:
            kwargs["project_id"] = None if project_id is None else str(project_id)
        response = cls._client.patch(
            cls._path.format(userBlueprintId=user_blueprint_id),
            data=dict(decompressed_blueprint=True, **kwargs),
            keep_attrs=["value", "project_id"],
        )
        return cls.from_server_data(response.json())

    @classmethod
    def delete(cls, user_blueprint_id: str) -> Response:
        """
        Delete a user blueprint, specified by the `userBlueprintId`.

        Parameters
        ----------
        user_blueprint_id: string
            Used to identify a specific user-owned blueprint.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        :class:`requests.models.Response`
        """
        return cast(
            Response, cls._client.delete(cls._path.format(userBlueprintId=user_blueprint_id))
        )

    @classmethod
    def get_input_types(cls) -> UserBlueprintAvailableInput:
        """
        Retrieve the input types which can be used with User Blueprints.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprintAvailableInput
        """
        return UserBlueprintAvailableInput.get_input_types()

    @classmethod
    def add_to_project(
        cls,
        project_id: str,
        user_blueprint_ids: Union[str, List[str]],
    ) -> UserBlueprintAddToProjectMenu:
        """
        Add a list of user blueprints, by id, to a specified (by id) project's repository.

        Parameters
        ----------
        project_id: string
            The projectId of the project for the repository to add the specified user blueprints
            to.
        user_blueprint_ids: list(string) or string
            The ids of the user blueprints to add to the specified project's repository.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprintAddToProjectMenu
        """
        user_blueprint_ids = (
            [str(u) for u in user_blueprint_ids]
            if isinstance(user_blueprint_ids, list)
            else [str(user_blueprint_ids)]
        )
        return UserBlueprintAddToProjectMenu.add_to_project(
            project_id=str(project_id), user_blueprint_ids=user_blueprint_ids
        )

    @classmethod
    def get_available_tasks(
        cls,
        project_id: Optional[str] = None,
        user_blueprint_id: Optional[str] = None,
    ) -> UserBlueprintAvailableTasks:
        """
        Retrieve the available tasks, organized into categories, which can be used to create or
        modify User Blueprints.

        Parameters
        ----------
        project_id: string, Optional
        user_blueprint_id: string, Optional

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprintAvailableTasks
        """
        return UserBlueprintAvailableTasks.get_available_tasks(
            project_id=None if project_id is None else str(project_id),
            user_blueprint_id=None if user_blueprint_id is None else str(user_blueprint_id),
        )

    @classmethod
    def validate_task_parameters(
        cls,
        output_method: str,
        task_code: str,
        task_parameters,
        project_id: Optional[str] = None,
    ) -> UserBlueprintValidateTaskParameters:
        """
        Validate that each value assigned to specified task parameters are valid.

        Parameters
        ----------
        output_method: enum('P', 'Pm', 'S', 'Sm', 'T', 'TS')
            The method representing how the task will output data.
        task_code: string
            The task code representing the task to validate parameter values.
        task_parameters: list(UserBlueprintTaskParameterValidationRequestParamItem)
            A list of task parameters and proposed values to be validated.
        project_id: string (optional, default is None)
            The projectId representing the project where this user blueprint is edited.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprintValidateTaskParameters
        """
        return UserBlueprintValidateTaskParameters.validate_task_parameters(
            output_method=output_method,
            task_code=task_code,
            task_parameters=task_parameters,
            project_id=project_id,
        )

    @classmethod
    def list_shared_roles(
        cls,
        user_blueprint_id: str,
        limit: int = 100,
        offset: int = 0,
        id: Optional[str] = None,
        name: Optional[str] = None,
        share_recipient_type: Optional[str] = None,
    ) -> List[UserBlueprintSharedRolesResponseValidator]:
        """
        Get a list of users, groups and organizations that have an access to this user blueprint

        Parameters
        ----------
        id: str, Optional
            Only return the access control information for a organization, group or user with this
            ID.
        limit: int (Default=100)
            At most this many results are returned.
        name: string, Optional
            Only return the access control information for a organization, group or user with this
            name.
        offset: int (Default=0)
            This many results will be skipped.
        share_recipient_type: enum('user', 'group', 'organization'), Optional
            Describes the recipient type, either user, group, or organization.
        user_blueprint_id: str
            Used to identify a specific user-owned blueprint.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        list[UserBlueprintSharedRolesResponseValidator]
        """
        return UserBlueprintSharingListController.list_shared_roles(
            user_blueprint_id=user_blueprint_id,
            limit=limit,
            offset=offset,
            id=id,
            name=name,
            share_recipient_type=share_recipient_type,
        )

    @classmethod
    def validate_blueprint(
        cls, blueprint, project_id: Optional[str] = None
    ) -> List[VertexContextItem]:
        """Validate a user blueprint and return information about the inputs expected and outputs
        provided by each task.

        Parameters
        ----------
        blueprint: list(dict) or list(UserBlueprintTask)
            The representation of a directed acyclic graph defining a pipeline of data through tasks
            and a final estimator.
        project_id: string (optional, default is None)
            The projectId representing the project where this user blueprint is edited.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        list[VertexContextItem]
        """
        return UserBlueprintValidate.validate_blueprint(blueprint, project_id=project_id)

    @classmethod
    def update_shared_roles(
        cls,
        user_blueprint_id: str,
        roles: List[
            Union[GrantAccessControlWithUsernameValidator, GrantAccessControlWithIdValidator]
        ],
    ) -> Response:
        """
        Share a user blueprint with a user, group, or organization

        Parameters
        ----------
        user_blueprint_id: str
            Used to identify a specific user-owned blueprint.
        roles: list(or(GrantAccessControlWithUsernameValidator, GrantAccessControlWithIdValidator))
            Array of GrantAccessControl objects., up to maximum 100 objects.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        :class:`requests.models.Response`
        """
        return UserBlueprintSharingUpdateController.update_shared_roles(
            user_blueprint_id=user_blueprint_id, roles=roles
        )

    @classmethod
    def search_catalog(
        cls,
        search: Optional[str] = None,
        tag: Optional[str] = None,
        limit: int = 100,
        offset: int = 0,
        owner_user_id: Optional[str] = None,
        owner_username: Optional[str] = None,
        order_by: str = "-created",
    ) -> UserBlueprintCatalogSearch:
        """
        Fetch a list of the user blueprint catalog entries the current user has access to
        based on an optional search term, tags, owner user info, or sort order.

        Parameters
        ----------
        search: string, Optional.
            A value to search for in the dataset's name, description, tags, column names,
            categories, and latest error. The search is case insensitive. If no value is provided
            for this parameter, or if the empty string is used, or if the string contains only
            whitespace, no filtering will be done. Partial matching is performed on dataset name
            and description fields while all other fields will only match if the search matches the
            whole value exactly.

        tag: string, Optional.
            If provided, the results will be filtered to include only items with the specified tag.

        limit: int, Optional. (default: 0), at most this many results are returned. To specify no
            limit, use 0. The default may change and a maximum limit may be imposed without notice.

        offset: int, Optional. (default: 0), this many results will be skipped.

        owner_user_id: string, Optional.
            Filter results to those owned by one or more owner identified by UID.

        owner_username: string,  Optional.
            Filter results to those owned by one or more owner identified by username.

        order_by: string, Optional. Defaults to '-created'.
            Sort order which will be applied to catalog list, valid options are "catalogName",
            "originalName", "description", "created", and "relevance". For all options other
            than relevance, you may prefix the attribute name with a dash to sort
            in descending order. e.g. orderBy='-catalogName'.
        """
        return UserBlueprintCatalogSearch.search_catalog(
            search=search,
            tag=tag,
            limit=limit,
            offset=offset,
            owner_user_id=owner_user_id,
            owner_username=owner_username,
            order_by=order_by,
        )


class UserBlueprintCatalogSearch(APIObject, HumanReadable, CompareWithJSON):
    """
    An APIObject representing a user blueprint catalog entry the current
    user has access to based on an optional search term and/or tags.

    Parameters
    ----------
    id: str
        The ID of the catalog entry linked to the user blueprint.

    catalog_name: str
        The name of the user blueprint.

    creator: str
        The name of the user that created the user blueprint.

    user_blueprint_id: str
        The ID of the user blueprint.

    description: str, Optional (Default=None)
        The description of the user blueprint.

    last_modifier_name: str, Optional (Default=None)
        The name of the user that last modified the user blueprint.
    """

    _path = "catalogItems/"
    _converter = UserBlueprintCatalogSearchItem_

    def __init__(
        self,
        id: str,
        catalog_name: str,
        info_creator_full_name: str,
        user_blueprint_id: str,
        description: Optional[str] = None,
        last_modifier_full_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self.id = id
        self.catalog_name = catalog_name
        self.creator = info_creator_full_name
        self.user_blueprint_id = user_blueprint_id
        self.description = description
        self.last_modifier_name = last_modifier_full_name

    @classmethod
    def search_catalog(
        cls,
        search: Optional[str] = None,
        tag: Optional[str] = None,
        limit: int = 100,
        offset: int = 0,
        owner_user_id: Optional[str] = None,
        owner_username: Optional[str] = None,
        order_by: str = "-created",
    ) -> List[UserBlueprintCatalogSearch]:
        """
        Fetch a list of the user blueprint catalog entries the current user has access to
        based on an optional search term, tags, owner user info, or sort order.

        Parameters
        ----------
        search: string, Optional.
            A value to search for in the dataset's name, description, tags, column names,
            categories, and latest error. The search is case insensitive. If no value is provided
            for this parameter, or if the empty string is used, or if the string contains only
            whitespace, no filtering will be done. Partial matching is performed on dataset name
            and description fields while all other fields will only match if the search matches the
            whole value exactly.

        tag: string, Optional.
            If provided, the results will be filtered to include only items with the specified tag.

        limit: int, Optional. (default: 0), at most this many results are returned. To specify no
            limit, use 0. The default may change and a maximum limit may be imposed without notice.

        offset: int, Optional. (default: 0), this many results will be skipped.

        owner_user_id: string, Optional.
            Filter results to those owned by one or more owner identified by UID.

        owner_username: string,  Optional.
            Filter results to those owned by one or more owner identified by username.

        order_by: string, Optional. Defaults to '-created'.
            Sort order which will be applied to catalog list, valid options are "catalogName",
            "originalName", "description", "created", and "relevance". For all options other
            than relevance, you may prefix the attribute name with a dash to sort
            in descending order. e.g. orderBy='-catalogName'.
        """
        params = dict(
            search_for=search,
            tag=tag,
            type="user_blueprint",
            limit=limit or DEFAULT_BATCH_SIZE,
            offset=offset,
            owner_user_id=owner_user_id,
            owner_username=owner_username,
            order_by=order_by,
        )
        if limit == 0:
            data = list(unpaginate(cls._path, params, cls._client))
        else:
            data = cls._client.get(cls._path, params=params).json()["data"]

        return [cls.from_server_data(a) for a in data]


class UserBlueprintAvailableInput(APIObject, HumanReadable, CompareWithJSON):
    """
    Retrieve the input types which can be used with User Blueprints.

    Parameters
    ----------
    input_types: list(UserBlueprintsInputType)
        A list of associated pairs of an input types and their human-readable names.
    """

    _path = "userBlueprintsInputTypes/"
    _converter = UserBlueprintsInputTypesResponse_

    def __init__(self, input_types: List[UserBlueprintsInputType], **kwargs: Any) -> None:
        self.input_types = _init_list_of_class_or_dict(input_types, UserBlueprintsInputType)

    @classmethod
    def get_input_types(cls) -> UserBlueprintAvailableInput:
        """
        Retrieve the input types which can be used with User Blueprints.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprintAvailableInput
        """
        response = cls._client.get(cls._path)
        return cls.from_server_data(response.json())


class UserBlueprintAddToProjectMenu(APIObject, HumanReadable, CompareWithJSON):
    """
    Add a list of user blueprints, by id, to a specified (by id) project's repository.

    Parameters
    ----------
    added_to_menu: list(UserBlueprintAddedToMenuItem)
        The list of userBlueprintId and blueprintId pairs representing blueprints successfully
        added to the project repository.
    not_added_to_menu: list(UserBlueprintNotAddedToMenuItem)
        The list of userBlueprintId and error message representing blueprints which failed to be
        added to the project repository.
    message: string
        A success message or a list of reasons why the list of blueprints could not be added
        to the project repository.
    """

    _path = "userBlueprintsProjectBlueprints/"
    _converter = UserBlueprintAddToMenuResponse_

    def __init__(
        self,
        added_to_menu: List[UserBlueprintAddedToMenuItem],
        not_added_to_menu: List[UserBlueprintAddedToMenuItem] = None,
        message: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self.added_to_menu = _init_list_of_class_or_dict(
            added_to_menu, UserBlueprintAddedToMenuItem
        )
        self.not_added_to_menu = _init_list_of_class_or_dict(
            not_added_to_menu, UserBlueprintNotAddedToMenuItem
        )
        self.message = message

    @classmethod
    def add_to_project(
        cls, project_id: str, user_blueprint_ids: List[str]
    ) -> UserBlueprintAddToProjectMenu:
        """
        Add a list of user blueprints, by id, to a specified (by id) project's repository.

        Parameters
        ----------
        project_id: string
            The projectId of the project for the repository to add the specified user blueprints
            to.
        user_blueprint_ids: list(string)
            The ids of the user blueprints to add to the specified project's repository.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprintAddToProjectMenu
        """
        response = cls._client.post(
            cls._path,
            data=dict(
                project_id=str(project_id), user_blueprint_ids=[str(u) for u in user_blueprint_ids]
            ),
        )
        return cls.from_server_data(response.json())


class UserBlueprintValidateTaskParameters(APIObject, HumanReadable, CompareWithJSON):
    """
    Validate that each value assigned to specified task parameters are valid.

    Parameters
    ----------
    errors: list(UserBlueprintsValidateTaskParameter)
        A list of the task parameters, their proposed values, and messages describing why each is
        not valid.
    """

    _path = "userBlueprintsTaskParameters/"
    _converter = UserBlueprintsValidateTaskParametersResponse_

    def __init__(self, errors: List[UserBlueprintsValidateTaskParameter], **kwargs: Any) -> None:
        self.errors = _init_list_of_class_or_dict(errors, UserBlueprintsValidateTaskParameter)

    @classmethod
    def validate_task_parameters(
        cls,
        output_method: str,
        task_code: str,
        task_parameters: List[UserBlueprintTaskParameterValidationRequestParamItem],
        project_id: Optional[str] = None,
    ) -> UserBlueprintValidateTaskParameters:
        """
        Validate that each value assigned to specified task parameters are valid.

        Parameters
        ----------
        output_method: enum('P', 'Pm', 'S', 'Sm', 'T', 'TS')
            The method representing how the task will output data.
        task_code: string
            The task code representing the task to validate parameter values.
        task_parameters: list(UserBlueprintTaskParameterValidationRequestParamItem)
            A list of task parameters and proposed values to be validated.
        project_id: string (optional, default is None)
            The projectId representing the project where this user blueprint is edited.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprintValidateTaskParameters
        """
        response = cls._client.post(
            cls._path,
            data=dict(
                task_code=task_code,
                output_method=output_method,
                task_parameters=[
                    t.as_json() if not isinstance(t, dict) else t for t in task_parameters
                ],
                project_id=project_id,
            ),
        )
        return cls.from_server_data(response.json())


class UserBlueprintValidate(APIObject, HumanReadable, CompareWithJSON):
    """
    Validate the structure of the blueprint and that blueprint inputs/outputs are valid. Also return
    information about task data inputs and outputs.

    Parameters
    ----------
    vertex_context: list(VertexContextItem)
        Information about blueprint vertices, including warning and error messages.
    blueprint_context: VertexContextItemMessages
        Warnings and errors which may describe or summarize warnings or errors in the blueprint's
        vertices
    """

    _path = "userBlueprintsValidations/"
    _converter = UserBlueprintValidationResponse_

    def __init__(
        self,
        vertex_context: List[VertexContextItem],
        blueprint_context: VertexContextItemMessages,
        **kwargs: Any,
    ) -> None:
        self.blueprint_context = _init_class_or_dict(blueprint_context, VertexContextItemMessages)
        self.vertex_context = _init_list_of_class_or_dict(vertex_context, VertexContextItem)

    @classmethod
    def validate_blueprint(
        cls,
        blueprint,
        project_id: Optional[str] = None,
    ) -> UserBlueprintValidate:
        """Validate the structure of the blueprint and that blueprint inputs/outputs are valid.
        Also return information about task data inputs and outputs.
        Parameters
        ----------
        blueprint: list(dict) or list(UserBlueprintTask)
            The representation of a directed acyclic graph defining a pipeline of data through tasks
            and a final estimator.
        project_id: string, Optional (Default=None)
            The project associated with the blueprint. Necessary in the event of project specific
            tasks, such as column selection tasks.
        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        Returns
        -------
        UserBlueprintValidate
        """
        response = cls._client.post(
            cls._path,
            data=dict(
                blueprint=[b.as_json() if not isinstance(b, dict) else b for b in blueprint],
                project_id=project_id,
            ),
        )
        return cls.from_server_data(response.json())


class UserBlueprintAvailableTasks(APIObject, HumanReadable, CompareWithJSON):
    """
    Retrieve the available tasks, organized into categories, which can be used to create or modify
    User Blueprints.

    Parameters
    ----------
    categories: list(UserBlueprintTaskCategoryItem)
        A list of the available task categories, sub-categories, and tasks.
    tasks: list(UserBlueprintTaskLookupEntry)
        A list of task codes and their task definitions.
    """

    _path = "userBlueprintsTasks/"
    _converter = UserBlueprintTasksResponse_

    def __init__(
        self,
        categories: List[UserBlueprintTaskCategoryItem],
        tasks: List[UserBlueprintTaskLookupEntry],
        **kwargs: Any,
    ) -> None:
        self.categories = _init_list_of_class_or_dict(categories, UserBlueprintTaskCategoryItem)
        self.tasks = _init_list_of_class_or_dict(tasks, UserBlueprintTaskLookupEntry)

    @classmethod
    def get_available_tasks(
        cls, project_id: Optional[str] = None, user_blueprint_id: Optional[str] = None
    ) -> UserBlueprintAvailableTasks:
        """
        Retrieve the available tasks, organized into categories, which can be used to create or
        modify User Blueprints.

        Parameters
        ----------
        project_id: string, Optional
        user_blueprint_id: string, Optional

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        UserBlueprintAvailableTasks
        """
        response = cls._client.get(
            cls._path,
            params=dict(
                project_id=None if project_id is None else str(project_id),
                user_blueprint_id=None if user_blueprint_id is None else str(user_blueprint_id),
            ),
        )
        return cls.from_server_data(response.json())


class UserBlueprintValidationRequest(HumanReadable, CompareWithJSON):
    """
    Blueprint and optional project id for validation.

    Attributes
    ----------
    blueprint: list(dict) or list(UserBlueprintTask)
         The blueprint in JSON format.
    project_id: string
         The project associated with the blueprint. Necessary for
         project-specific tasks, such as column selection tasks.
    """

    def __init__(
        self,
        blueprint,
        project_id: Optional[str] = None,
    ) -> None:
        self.project_id = project_id
        self.blueprint = blueprint


class UserBlueprintValidationResponse(HumanReadable, CompareWithJSON):
    """
    Vertex context / validation information and warnings or errors.

    Attributes
    ----------
    vertex_context: list(VertexContextItem)
         Info, warnings, and errors about each vertex in the blueprint.
    blueprint_context: VertexContextItemMessages
        Warnings and errors which may describe or summarize warnings or errors in the blueprint's
        vertices
    """

    def __init__(
        self,
        vertex_context: List[VertexContextItem],
        blueprint_context: VertexContextItemMessages,
    ) -> None:
        self.vertex_context = vertex_context
        self.blueprint_context = blueprint_context


class UserBlueprintSharingListController(APIObject, HumanReadable, CompareWithJSON):
    """
    Get a list of users, groups and organizations that have an access to this user blueprint

    Parameters
    ----------
    count: int, Optional
        Number of items returned on this page.
    next: url or null
        URL pointing to the next page (if null, there is no next page).
    previous: url or null
        URL pointing to the previous page (if null, there is no previous page).
    data: list(UserBlueprintSharedRolesResponseValidator)
        A list of SharedRoles objects.
    total_count: int
        The total number of items across all pages.
    """

    _path = "userBlueprints/{userBlueprintId}/sharedRoles/"
    _converter = UserBlueprintSharedRolesListResponseValidator_

    def __init__(
        self,
        data: List[UserBlueprintSharedRolesResponseValidator],
        **kwargs: Any,
    ) -> None:
        self.data = _init_list_of_class_or_dict(data, UserBlueprintSharedRolesResponseValidator)

    @classmethod
    def list_shared_roles(
        cls,
        user_blueprint_id: str,
        limit: int = 100,
        offset: int = 0,
        id: Optional[str] = None,
        name: Optional[str] = None,
        share_recipient_type: Optional[str] = None,
    ) -> List[UserBlueprintSharedRolesResponseValidator]:
        """
        Get a list of users, groups and organizations that have an access to this user blueprint

        Parameters
        ----------
        id: str, Optional
            Only return the access control information for a organization, group or user with this
            ID.
        limit: int (Default=100)
            At most this many results are returned.
        name: string, Optional
            Only return the access control information for a organization, group or user with this
            name.
        offset: int (Default=0)
            This many results will be skipped.
        share_recipient_type: enum('user', 'group', 'organization'), Optional
            Describes the recipient type, either user, group, or organization.
        user_blueprint_id: str
            Used to identify a specific user-owned blueprint.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        list(UserBlueprintSharedRolesResponseValidator)
        """
        path = cls._path.format(userBlueprintId=user_blueprint_id)
        params = dict(
            offset=offset,
            limit=limit or DEFAULT_BATCH_SIZE,
            name=name,
            id=id,
            share_recipient_type=share_recipient_type,
        )
        if limit == 0:
            data = list(unpaginate(path, params, cls._client))
        else:
            data = cls._client.get(path, params=params).json()["data"]
        return cls.from_server_data(dict(data=data)).data


class UserBlueprintSharingUpdateController(
    APIObject, HumanReadable, CompareWithJSON
):  # pylint: disable=missing-class-docstring
    _path = "userBlueprints/{userBlueprintId}/sharedRoles/"

    @classmethod
    def update_shared_roles(
        cls,
        user_blueprint_id: str,
        roles: List[
            Union[GrantAccessControlWithUsernameValidator, GrantAccessControlWithIdValidator]
        ],
    ) -> Response:
        """
        Share a user blueprint with a user, group, or organization

        Parameters
        ----------
        user_blueprint_id: str
            Used to identify a specific user-owned blueprint.
        roles: list(or(GrantAccessControlWithUsernameValidator, GrantAccessControlWithIdValidator))
            Array of GrantAccessControl objects., up to maximum 100 objects.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Returns
        -------
        :class:`requests.models.Response`
        """
        roles_json = []
        for role in roles:
            if not isinstance(
                role,
                (dict, GrantAccessControlWithUsernameValidator, GrantAccessControlWithIdValidator),
            ):
                raise TypeError(
                    "'roles' must be a list of one of: dict, "
                    "GrantAccessControlWithUsernameValidator, GrantAccessControlWithIdValidator"
                )
            roles_json.append(role if isinstance(role, dict) else role.as_json())
        return cast(
            Response,
            cls._client.patch(
                cls._path.format(userBlueprintId=user_blueprint_id),
                data=dict(operation="updateRoles", roles=roles_json),
            ),
        )


class UserBlueprintSharedRolesResponseValidator(HumanReadable, CompareWithJSON):
    """
    A list of SharedRoles objects.

    Parameters
    ----------
    share_recipient_type: enum('user', 'group', 'organization')
        Describes the recipient type, either user, group, or organization.
    role: str, one of enum('CONSUMER', 'EDITOR', 'OWNER')
        The role of the org/group/user on this dataset or "NO_ROLE" for removing access when used
        with route to modify access.
    id: str
        The ID of the recipient organization, group or user.
    name: string
        The name of the recipient organization, group or user.
    """

    def __init__(self, id: str, name: str, role, share_recipient_type: str, **kwargs: Any) -> None:
        self.share_recipient_type = share_recipient_type
        self.role = role
        self.id = id
        self.name = name


class GrantAccessControlWithIdValidator(HumanReadable, CompareWithJSON):
    """

    Parameters
    ----------
    share_recipient_type: enum('user', 'group', 'organization')
        Describes the recipient type, either user, group, or organization.
    role: string
        The role of the recipient on this entity. One of OWNER, USER, OBSERVER.
    id: str
        The ID of the recipient.
    """

    def __init__(
        self,
        id: str,
        role: str,
        share_recipient_type: str,
        **kwargs: Any,
    ) -> None:
        self.share_recipient_type = share_recipient_type
        self.role = role
        self.id = id


class GrantAccessControlWithUsernameValidator(HumanReadable, CompareWithJSON):
    """

    Parameters
    ----------
    share_recipient_type: enum('user', 'group', 'organization')
        Describes the recipient type, either user, group, or organization.
    role: string
        The role of the recipient on this entity. One of OWNER, USER, OBSERVER.
    username: string
        Username of the user to update the access role for.
    """

    def __init__(
        self,
        role: str,
        share_recipient_type: str,
        username: str,
        **kwargs: Any,
    ) -> None:
        self.share_recipient_type = share_recipient_type
        self.role = role
        self.username = username


class UserBlueprintHexColumnNameLookupEntry(HumanReadable, CompareWithJSON):
    """
    A lookup between hex values and data column names used in the blueprint.

    Parameters
    ----------
    colname: string
        The name of the column.
    hex: string
        A safe hex representation of the column name.
    project_id: string, Optional
        The id of the project from which the column name originates.
    """

    def __init__(
        self, colname: str, hex: str, project_id: Optional[str] = None, **kwargs: Any
    ) -> None:
        self.colname = colname
        self.hex = hex
        self.project_id = project_id


class UserBlueprintTask(HumanReadable, CompareWithJSON):
    """
    A task (i.e. vertex) in a blueprint

    Parameters
    ----------
    task_id: string
        The identifier of a task / vertex in the blueprint.
    task_data: UserBlueprintTaskData
        The data defining the task / vertex in the blueprint.
    """

    def __init__(self, task_data: UserBlueprintTaskData, task_id: str, **kwargs: Any) -> None:
        self.task_id = task_id
        self.task_data = _init_class_or_dict(task_data, UserBlueprintTaskData)


class UserBlueprintTaskData(HumanReadable, CompareWithJSON):
    """
    The data defining the task / vertex in the blueprint.

    Parameters
    ----------
    inputs: list(string)
        The ids or input data types which will be inputs to the task.
    output_method: string
        The method which the task will use to produce output.
    output_method_parameters: list(ParamValuePair)
        The parameters which further define how output will be produced.
    task_code: string
        The unique code representing the python class which will be instantiated and executed.
    task_parameters: list(ParamValuePair)
        The parameters which further define the behavior of the task.
    x_transformations: list(ParamValuePair)
        Transformations to apply to the input data before fitting or predicting.
    y_transformations: list(ParamValuePair)
        Transformations to apply to the input target before fitting or predicting.
    """

    def __init__(
        self,
        inputs: List[str],
        output_method: str,
        output_method_parameters: List[ParamValuePair],
        task_code: str,
        task_parameters: List[ParamValuePair],
        x_transformations: List[ParamValuePair],
        y_transformations: List[ParamValuePair],
        **kwargs: Any,
    ) -> None:
        self.inputs = inputs
        self.output_method = output_method
        self.output_method_parameters = _init_list_of_class_or_dict(
            output_method_parameters, ParamValuePair
        )
        self.task_code = task_code
        self.task_parameters = _init_list_of_class_or_dict(task_parameters, ParamValuePair)
        self.x_transformations = _init_list_of_class_or_dict(x_transformations, ParamValuePair)
        self.y_transformations = _init_list_of_class_or_dict(y_transformations, ParamValuePair)


class VertexContextItem(HumanReadable, CompareWithJSON):
    """
    Info about, warnings about, and errors with a specific vertex in the blueprint.

    Parameters
    ----------
    task_id: string
        The id associated with a specific vertex in the blueprint.
    information: VertexContextItemInfo
    messages: VertexContextItemMessages
    """

    def __init__(
        self,
        information: VertexContextItemInfo,
        messages: VertexContextItemMessages,
        task_id: str,
        **kwargs: Any,
    ) -> None:
        """
        Parameters
        ----------
        task_id: string
            The id associated with a specific vertex in the blueprint.
        information: dict
        messages: dict
        """
        self.task_id = task_id
        self.information = _init_class_or_dict(information, VertexContextItemInfo)
        self.messages = _init_class_or_dict(messages, VertexContextItemMessages)


class VertexContextItemMessages(HumanReadable, CompareWithJSON):
    """

    Parameters
    ----------
    errors: list(string), Optional
        Errors with a specific vertex in the blueprint. Execution of the vertex is expected to
        fail.
    warnings: list(string), Optional
        Warnings about a specific vertex in the blueprint. Execution of the vertex may fail or
        behave unexpectedly.
    """

    def __init__(
        self,
        errors: Optional[List[str]] = None,
        warnings: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        self.errors = errors
        self.warnings = warnings


class VertexContextItemInfo(HumanReadable, CompareWithJSON):
    """

    Parameters
    ----------
    inputs: list(string)
        A specification of requirements of the inputs of the vertex.
    outputs: list(string)
        A specification of expectations of the output of the vertex.
    """

    def __init__(self, inputs: List[str], outputs: List[str], **kwargs: Any) -> None:
        self.inputs = inputs
        self.outputs = outputs


class ParamValuePair(CompareWithJSON):
    """
    Transformations to apply to the input target before fitting or predicting.

    Parameters
    ----------
    param: string
        The name of a field associated with the value.
    value: any or list(any)
        Any value.
    """

    def __init__(
        self, param: str, value: Optional[Union[Any, List[Any]]] = None, **kwargs: Any
    ) -> None:
        self.param = param
        self.value = value

    def __repr__(self) -> str:
        return f"({self.param}={self.value})"


class UserBlueprintsInputType(HumanReadable, CompareWithJSON):
    """
    A list of associated pairs of an input types and their human-readable names.

    Parameters
    ----------
    type: string
        The unique identifier of an input type.
    name: string
        The human-readable name of an input type.
    """

    def __init__(self, name: str, type: str, **kwargs: Any) -> None:
        self.type = type
        self.name = name


class UserBlueprintAddedToMenuItem(HumanReadable, CompareWithJSON):
    """
    A userBlueprintId and blueprintId pair representing blueprints successfully added to
    the project repository.

    Parameters
    ----------
    blueprint_id: string
        The blueprintId representing the blueprint which was added to the project repository.
    user_blueprint_id: string
        The userBlueprintId associated with the blueprintId added to the project repository.
    """

    def __init__(self, blueprint_id: str, user_blueprint_id: str, **kwargs: Any) -> None:
        self.blueprint_id = blueprint_id
        self.user_blueprint_id = user_blueprint_id


class UserBlueprintNotAddedToMenuItem(HumanReadable, CompareWithJSON):
    """
    A userBlueprintId and error pair representing blueprints which failed to be added to
    the project repository.

    Parameters
    ----------
    error: string
        The error message representing why the blueprint was not added to the project repository.
    user_blueprint_id: string
        The userBlueprintId associated with the blueprint which was not added to the project
        repository.
    """

    def __init__(self, blueprint_id: str, user_blueprint_id: str, **kwargs: Any) -> None:
        self.blueprint_id = blueprint_id
        self.user_blueprint_id = user_blueprint_id


class UserBlueprintsValidateTaskParameter(HumanReadable, CompareWithJSON):
    """
    A list of the task parameters, their proposed values, and messages describing why each is not
    valid.

    Parameters
    ----------
    message: string
        The description of the issue with the proposed task parameter value.
    param_name: string
        The name of the validated task parameter.
    value: any or list(any)
        The invalid value proposed for the validated task parameter.
    """

    def __init__(
        self,
        message: str,
        param_name: str,
        value: Optional[Union[Any, List[Any]]] = None,
        **kwargs: Any,
    ) -> None:
        self.message = message
        self.param_name = param_name
        self.value = value


class UserBlueprintTaskParameterValidationRequestParamItem(HumanReadable, CompareWithJSON):
    """
    A list of task parameters and proposed values to be validated.

    Parameters
    ----------
    param_name: string
        The name of the task parameter to be validated.
    new_value: any or list(any)
        The proposed value for the task parameter.
    """

    def __init__(
        self,
        param_name: str,
        new_value: Optional[Union[Any, List[Any]]] = None,
        **kwargs: Any,
    ) -> None:
        self.param_name = param_name
        self.new_value = new_value


class UserBlueprintTaskLookupEntry(HumanReadable, CompareWithJSON):
    """
    A list of task codes and their task definitions.

    Parameters
    ----------
    task_code: string
        The unique code which represents the task to be constructed and executed
    task_definition: UserBlueprintTaskDefinition
        A definition of a task in terms of label, arguments, description, and other metadata.
    """

    def __init__(
        self, task_code: str, task_definition: UserBlueprintTaskDefinition, **kwargs: Any
    ) -> None:
        self.task_code = task_code
        self.task_definition = _init_class_or_dict(task_definition, UserBlueprintTaskDefinition)


class UserBlueprintTaskDefinition(HumanReadable, CompareWithJSON):
    """
    A definition of a task in terms of label, arguments, description, and other metadata.

    Parameters
    ----------
    task_code: string
        The unique code which represents the task to be constructed and executed
    label: string
        The generic / default title or label for the task.
    description: string
        A description of the task.
    arguments: list(UserBlueprintTaskArgument)
        A list of definitions of each argument which can be set for the task.
    categories: list(string)
        The categories which the task is in.
    colnames_and_types: list(ColnameAndType), Optional
        The column names, their types, and their hex representation, available in the specified
        project for the task.
    icon: int
        The integer representing the id to be displayed when the blueprint is trained.
    output_methods: list(string)
        The methods which the task can use to produce output.
    time_series_only: bool
        Whether the task can only be used with time series projects.
    url: dict or string or TaskDocumentationUrl
        The URL of the documentation of the task.
    valid_inputs: list(string)
    is_custom_task: bool, Optional
        Whether the task is custom code written by the user.
    custom_task_id: string or null, Optional
        The id of the custom task, if it is a custom task.
    custom_task_versions: list(UserBlueprintTaskCustomTaskMetadata), Optional
        Metadata for all of the custom task's versions.
    supports_scoring_code: bool
        Whether Scoring Code can be generated for the task.
    """

    def __init__(
        self,
        arguments: List[UserBlueprintTaskArgument],
        categories: List[str],
        description: str,
        icon: int,
        label: str,
        output_methods: List[str],
        task_code: str,
        time_series_only: bool,
        url,
        valid_inputs: List[str],
        colnames_and_types: Optional[List[ColnameAndType]] = None,
        custom_task_id: Optional[str] = None,
        custom_task_versions: Optional[List[UserBlueprintTaskCustomTaskMetadata]] = None,
        is_custom_task: Optional[bool] = None,
        supports_scoring_code: bool = False,
        **kwargs: Any,
    ) -> None:
        self.task_code = task_code
        self.label = label
        self.description = description
        self.arguments = _init_list_of_class_or_dict(arguments, UserBlueprintTaskArgument)
        self.categories = categories
        self.colnames_and_types = _init_list_of_class_or_dict(colnames_and_types, ColnameAndType)
        self.icon = icon
        self.output_methods = output_methods
        self.time_series_only = time_series_only
        self.url = _init_class_or_dict(url, TaskDocumentationUrl)
        self.valid_inputs = valid_inputs
        self.is_custom_task = is_custom_task
        self.custom_task_id = custom_task_id
        self.custom_task_versions = _init_list_of_class_or_dict(
            custom_task_versions, UserBlueprintTaskCustomTaskMetadata
        )
        self.supports_scoring_code = supports_scoring_code


class UserBlueprintTaskCustomTaskMetadata(HumanReadable, CompareWithJSON):
    """
    Metadata for all of the custom task's versions.

    Parameters
    ----------
    id: string
        Id of the custom task version. The ID can be latest_<task_id> which implies to use the
        latest version of that custom task.
    version_major: int
        Major version of the custom task.
    version_minor: int
        Minor version of the custom task.
    label: string
        The name of the custom task version.
    """

    def __init__(
        self, id: str, label: str, version_major: int, version_minor: int, **kwargs: Any
    ) -> None:
        self.id = id
        self.version_major = version_major
        self.version_minor = version_minor
        self.label = label


class TaskDocumentationUrl(HumanReadable, CompareWithJSON):
    """
    The url of the documentation, if one exists.

    Parameters
    ----------
    documentation: string, Optional
        The URL of the documentation of the task.
    """

    def __init__(self, documentation: Optional[str] = None, **kwargs: Any) -> None:
        self.documentation = documentation
        if self.documentation is None:
            self.documentation = kwargs.get("data_robot _model _docs")


class ColnameAndType(HumanReadable, CompareWithJSON):
    """
    The column names, their types, and their hex representation, available in the specified project
    for the task.

    Parameters
    ----------
    hex: string
        A safe hex representation of the column name.
    colname: string
        The column name.
    type: string
        The data type of the column.
    """

    def __init__(self, colname: str, hex: str, type: str, **kwargs: Any) -> None:
        self.hex = hex
        self.colname = colname
        self.type = type


class UserBlueprintTaskArgument(HumanReadable, CompareWithJSON):
    """
    A list of definitions of each argument which can be set for the task.

    Parameters
    ----------
    key: string
        The unique key of the argument
    argument: UserBlueprintTaskArgumentDefinition
        The definition of a task argument, used to specify a certain aspect of the task.
    """

    def __init__(
        self, argument: UserBlueprintTaskArgumentDefinition, key: str, **kwargs: Any
    ) -> None:
        self.key = key
        self.argument = _init_class_or_dict(argument, UserBlueprintTaskArgumentDefinition)


class UserBlueprintTaskArgumentDefinition(HumanReadable, CompareWithJSON):
    """
    The definition of a task argument, used to specify a certain aspect of the task.

    Parameters
    ----------
    name: string
        The name of the argument.
    type: string
        The type of the argument (e.g. "int", "float", "select", "intgrid", "multi", etc.)
    default: any or list(any), Optional
        The default value of the argument.
    values: any or list(any) or dict
        The possible values of the argument, which may be a range or list, or dictionary of ranges
        or lists keyed by type.
    tunable: bool, Optional
        Whether the argument is tunable by the end-user.
    recommended: any or list(any), Optional
        The recommended value, based on frequently used values.
    """

    def __init__(
        self,
        name: str,
        type: str,
        values,
        default: Optional[Union[Any, List[Any]]] = None,
        recommended: Optional[Union[Any, List[Any]]] = None,
        tunable: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        self.name = name
        self.type = type
        self.default = default
        self.values = values
        self.tunable = tunable
        self.recommended = recommended


class UserBlueprintTaskCategoryItem(HumanReadable, CompareWithJSON):
    """
    A list of the available task categories, sub-categories, and tasks.

    Parameters
    ----------
    name: string
        The name of the category.
    task_codes: list(string)
        A list of task codes representing the tasks in this category.
    subcategories: list(UserBlueprintTaskCategoryItem), Optional
        A list of the available task category items.
    """

    def __init__(
        self,
        name: str,
        task_codes: List[str],
        subcategories: Optional[List[UserBlueprintTaskCategoryItem]] = None,
        **kwargs: Any,
    ) -> None:
        self.name = name
        self.task_codes = task_codes
        self.subcategories = _init_list_of_class_or_dict(
            subcategories, UserBlueprintTaskCategoryItem
        )


def _init_class_or_dict(item, class_):
    """
    Given None, a dictionary or an instance of `class_` return it
    as the instantiated form of `class_` or None.

    Parameters
    ----------
    item: dict or class_ instance or None
    class_: Type

    Returns
    -------
    class_ instance or None
    """
    return class_(**item) if isinstance(item, dict) else item


def _init_list_of_class_or_dict(items, class_):
    """
    Given a list of ( None, a dictionary or an instance of `class_` )
    return a list of their instantiated form of `class_` or as None.

    Parameters
    ----------
    items: list(dict or class_ instance or None)
    class_: Type

    Returns
    -------
    list(class_ instance) or None
    """
    return None if items is None else [class_(**a) if isinstance(a, dict) else a for a in items]



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import dateutil
import trafaret as t

from datarobot._compat import String
from datarobot.enums import (
    PredictionEnvironmentHealthType,
    PredictionEnvironmentModelFormats,
    PredictionEnvironmentPlatform,
)
from datarobot.models.api_object import APIObject
from datarobot.models.credential import Credential
from datarobot.models.data_store import DataStore
from datarobot.utils import pagination


class PredictionEnvironment(APIObject):
    """A prediction environment entity.

    .. versionadded:: v3.3.0

    Attributes
    ----------
    id: str
        The ID of the prediction environment.
    name: str
        The name of the prediction environment.
    description: str, optional
        The description of the prediction environment.
    platform: str, optional
        Indicates which platform is in use (AWS, GCP, DataRobot, etc.).
    permissions: list, optional
        A set of permissions for the prediction environment.
    is_deleted: boolean, optional
        The flag that shows if this prediction environment deleted.
    supported_model_formats: list[PredictionEnvironmentModelFormats], optional
        The list of supported model formats.
    is_managed_by_management_agent : boolean, optional
        Determines if the prediction environment should be managed by the management agent. False by default.
    datastore_id : str, optional
        The ID of the data store connection configuration.
        Only applicable for external prediction environments managed by DataRobot.
    credential_id : str, optional
        The ID of the credential associated with the data connection.
        Only applicable for external prediction environments managed by DataRobot.
    """

    _path = "predictionEnvironments/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("name"): String(),
            t.Key("platform"): t.Enum(*PredictionEnvironmentPlatform.ALL),
            t.Key("description", optional=True): String(allow_blank=True),
            t.Key("permissions", optional=True): t.List(t.String),
            t.Key("is_deleted"): t.Bool(),
            t.Key("supported_model_formats", optional=True): t.List(
                t.Enum(*PredictionEnvironmentModelFormats.ALL)
            ),
            t.Key("import_meta", optional=True): t.Dict(
                {
                    t.Key("date_created", optional=True): t.Call(dateutil.parser.parse),
                    t.Key("creator_id", optional=True): String(),
                    t.Key("creator_username", optional=True): String(),
                }
            ),
            t.Key("management_meta", optional=True): t.Dict(
                {
                    t.Key("user_id"): String(),
                    t.Key("additional_metadata", optional=True): t.List(
                        t.Dict({t.Key("key"): String(), t.Key("value"): String()})
                    ),
                    t.Key("username"): String(),
                }
            ),
            t.Key("health", optional=True): t.Dict(
                {
                    t.Key("status"): t.Enum(*PredictionEnvironmentHealthType.ALL),
                    t.Key("message"): String(),
                    t.Key("timestamp"): t.Call(dateutil.parser.parse),
                }
            ),
            t.Key("is_managed_by_management_agent", optional=True): t.Bool(),
            t.Key("plugin", optional=True): String(),
            t.Key("datastore_id", optional=True): String(),
            t.Key("credential_id", optional=True): String(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        name: str,
        platform: PredictionEnvironmentPlatform,
        description: Optional[str] = None,
        permissions: Optional[List[str]] = None,
        is_deleted: Optional[bool] = None,
        supported_model_formats: Optional[List[PredictionEnvironmentModelFormats]] = None,
        import_meta: Optional[Dict[str, Any]] = None,
        management_meta: Optional[Dict[str, Any]] = None,
        health: Optional[Dict[str, Any]] = None,
        is_managed_by_management_agent: Optional[bool] = None,
        plugin: Optional[str] = None,
        datastore_id: Optional[str] = None,
        credential_id: Optional[str] = None,
    ) -> None:
        self.id = id
        self.name = name
        self.platform = platform
        self.description = description
        self.plugin = plugin
        self.permissions = permissions
        self.is_deleted = is_deleted
        self.supported_model_formats = supported_model_formats
        self.import_meta = import_meta
        self.management_meta = management_meta
        self.health = health
        self.is_managed_by_management_agent = is_managed_by_management_agent
        self.datastore_id = datastore_id
        self.credential_id = credential_id

    def __repr__(self) -> str:
        return "{}('{}', '{}', '{}', '{}')".format(
            self.__class__.__name__,
            self.id,
            self.name,
            self.platform,
            self.description,
        )

    @classmethod
    def list(cls) -> List[PredictionEnvironment]:
        """
        Returns list of available external prediction environments.

        Returns
        -------
        prediction_environments : list of PredictionEnvironment instances
            contains a list of available prediction environments.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> prediction_environments = dr.PredictionEnvironment.list()
            >>> prediction_environments
            [
                PredictionEnvironment('5e429d6ecf8a5f36c5693e03', 'demo_pe', 'aws', 'env for demo testing'),
                PredictionEnvironment('5e42cc4dcf8a5f3256865840', 'azure_pe', 'azure', 'env for azure demo testing'),
            ]
        """

        return [
            cls.from_server_data(item) for item in pagination.unpaginate(cls._path, {}, cls._client)
        ]

    @classmethod
    def get(cls, pe_id: str) -> PredictionEnvironment:
        """
        Gets the PredictionEnvironment by id.

        Parameters
        ----------
        pe_id : str
            the identifier of the PredictionEnvironment.

        Returns
        -------
        prediction_environment : PredictionEnvironment
            the requested prediction environment object.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> pe = dr.PredictionEnvironment.get('5a8ac9ab07a57a1231be501f')
            >>> pe
            PredictionEnvironment('5a8ac9ab07a57a1231be501f', 'my_predict_env', 'aws', 'demo env'),
        """
        return cls.from_location(f"{cls._path}{pe_id}/")

    def delete(self) -> None:
        """
        Deletes the prediction environment.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> pe = dr.PredictionEnvironment.get('5a8ac9ab07a57a1231be501f')
            >>> pe.delete()
        """
        self._client.delete(f"{self._path}{self.id}/")

    @classmethod
    def create(
        cls,
        name: str,
        platform: PredictionEnvironmentPlatform,
        description: Optional[str] = None,
        plugin: Optional[str] = None,
        supported_model_formats: Optional[List[PredictionEnvironmentModelFormats]] = None,
        is_managed_by_management_agent: Optional[bool] = False,
        datastore: Optional[Union[DataStore, str]] = None,
        credential: Optional[Union[Credential, str]] = None,
    ) -> PredictionEnvironment:
        """
        Create a prediction environment.

        Parameters
        ----------
        name : str
            The name of the prediction environment.
        description : str, optional
            The description of the prediction environment.
        platform : str
            Indicates which platform is in use (AWS, GCP, DataRobot, etc.).
        plugin : str
            Optional. The plugin name to use.
        supported_model_formats : list[PredictionEnvironmentModelFormats], optional
            The list of supported model formats.
            When not provided, the default value is inferred based on platform, (DataRobot platform: DataRobot,
            Custom Models; All other platforms: DataRobot, Custom Models, External Models).
        is_managed_by_management_agent : boolean, optional
            Determines if this prediction environment should be managed by the management agent. default: False
        datastore : DataStore|str, optional]
            The datastore object or ID of the data store connection configuration.
            Only applicable for external Prediction Environments managed by DataRobot.
        credential : Credential|str, optional]
            The credential object or ID of the credential associated with the data connection.
            Only applicable for external Prediction Environments managed by DataRobot.

        Returns
        -------
        prediction_environment : PredictionEnvironment
            the prediction environment was created

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> pe = dr.PredictionEnvironment.create(
            ...     name='my_predict_env',
            ...     platform=PredictionEnvironmentPlatform.AWS,
            ...     description='demo prediction env',
            ... )
            >>> pe
            PredictionEnvironment('5e429d6ecf8a5f36c5693e99', 'my_predict_env', 'aws', 'demo prediction env'),
        """

        datastore_id = datastore
        credential_id = credential
        if isinstance(datastore, DataStore):
            datastore_id = datastore.id
        if isinstance(credential, Credential):
            credential_id = credential.credential_id

        payload = {
            "name": name,
            "platform": platform,
            "description": description,
            "plugin": plugin,
            "supportedModelFormats": supported_model_formats,
            "isManagedByManagementAgent": is_managed_by_management_agent,
            "datastoreId": datastore_id,
            "credentialId": credential_id,
        }
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.trafarets import UserBlueprintTaskArgument_

UserBlueprintsHexColumnNameLookupEntry_ = t.Dict(
    {
        t.Key("colname"): String(allow_blank=False),
        t.Key("hex"): String(allow_blank=False),
        t.Key("project_id", optional=True): String(allow_blank=False),
    }
)

ParamValuePair_ = t.Dict(
    {
        t.Key("param"): String(allow_blank=False),
        t.Key("value", optional=True): t.Or(
            t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
            t.List(t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
        ),
    }
)

UserBlueprintsBlueprintTaskData_ = t.Dict(
    {
        t.Key("inputs"): t.List(String(allow_blank=False)),
        t.Key("output_method"): String(allow_blank=False),
        t.Key("output_method_parameters"): t.List(ParamValuePair_),
        t.Key("task_code"): String(allow_blank=False),
        t.Key("task_parameters"): t.List(ParamValuePair_),
        t.Key("x_transformations"): t.List(ParamValuePair_),
        t.Key("y_transformations"): t.List(ParamValuePair_),
    }
)

UserBlueprintsBlueprintTask_ = t.Dict(
    {
        t.Key("task_id"): String(allow_blank=False),
        t.Key("task_data"): UserBlueprintsBlueprintTaskData_,
    }
)

VertexContextItemInfo_ = t.Dict(
    {
        t.Key("inputs"): t.List(String(allow_blank=False)),
        t.Key("outputs"): t.List(String(allow_blank=False)),
    }
)

VertexContextItemMessages_ = t.Dict(
    {
        t.Key("errors", optional=True): t.List(String(allow_blank=False)),
        t.Key("warnings", optional=True): t.List(String(allow_blank=False)),
    }
)

VertexContextItem_ = t.Dict(
    {
        t.Key("task_id"): String(allow_blank=False),
        t.Key("information"): t.Or(VertexContextItemInfo_),
        t.Key("messages"): t.Or(VertexContextItemMessages_),
    }
)

UserBlueprint_ = t.Dict(
    {
        t.Key("blender"): t.Bool(),
        t.Key("blueprint_id"): String(allow_blank=False),
        t.Key("custom_task_version_metadata", optional=True): t.List(
            t.List(String(allow_blank=False))
        ),
        t.Key("diagram"): String(allow_blank=False),
        t.Key("features"): t.List(String(allow_blank=False)),
        t.Key("features_text"): String(allow_blank=True),
        t.Key("hex_column_name_lookup", optional=True): t.List(
            UserBlueprintsHexColumnNameLookupEntry_
        ),
        t.Key("icons"): t.List(Int()),
        t.Key("insights"): String(allow_blank=False),
        t.Key("is_time_series", default=False): t.Bool(),
        t.Key("model_type"): String(allow_blank=False),
        t.Key("project_id", optional=True): String(allow_blank=False),
        t.Key("reference_model", default=False): t.Bool(),
        t.Key("shap_support", default=False): t.Bool(),
        t.Key("supported_target_types"): t.List(String(allow_blank=False)),
        t.Key("supports_gpu", default=False): t.Bool(),
        t.Key("user_blueprint_id"): String(allow_blank=False),
        t.Key("user_id"): String(allow_blank=False),
        t.Key("blueprint", optional=True): t.List(UserBlueprintsBlueprintTask_),
        t.Key("vertex_context", optional=True): t.List(VertexContextItem_),
        t.Key("blueprint_context", optional=True): VertexContextItemMessages_,
    }
).allow_extra("*")


UserBlueprintsInputType_ = t.Dict(
    {t.Key("type"): String(allow_blank=False), t.Key("name"): String(allow_blank=False)}
)

UserBlueprintsInputTypesResponse_ = t.Dict({t.Key("input_types"): t.List(UserBlueprintsInputType_)})


UserBlueprintAddedToMenuItem_ = t.Dict(
    {
        t.Key("blueprint_id"): String(allow_blank=False),
        t.Key("user_blueprint_id"): String(allow_blank=False),
    }
)

UserBlueprintNotAddedToMenuItem_ = t.Dict(
    {
        t.Key("error"): String(allow_blank=False),
        t.Key("user_blueprint_id"): String(allow_blank=False),
    }
)

UserBlueprintAddToMenuResponse_ = t.Dict(
    {
        t.Key("added_to_menu"): t.List(UserBlueprintAddedToMenuItem_),
        t.Key("not_added_to_menu", optional=True): t.List(UserBlueprintNotAddedToMenuItem_),
        t.Key("message", optional=True): String(allow_blank=False),
    }
).allow_extra("*")


UserBlueprintsValidateTaskParameter_ = t.Dict(
    {
        t.Key("message"): String(allow_blank=False),
        t.Key("param_name"): String(allow_blank=False),
        t.Key("value", optional=True): t.Or(
            t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
            t.List(t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
        ),
    }
)

UserBlueprintsValidateTaskParametersResponse_ = t.Dict(
    {t.Key("errors"): t.List(UserBlueprintsValidateTaskParameter_)}
)


UserBlueprintTaskCategoryItem_ = t.Dict(
    {
        t.Key("name"): String(allow_blank=False),
        t.Key("task_codes"): t.List(String(allow_blank=False)),
        t.Key("subcategories", optional=True): t.List(t.Dict().allow_extra("*")),
    }
).allow_extra("*")


ColnameAndType_ = t.Dict(
    {
        t.Key("hex"): String(allow_blank=False),
        t.Key("colname"): String(allow_blank=False),
        t.Key("type"): String(allow_blank=False),
    }
)

TaskDocumentationUrl_ = t.Dict(
    {t.Key("documentation", optional=True): String(allow_blank=True)}
).allow_extra("*")

UserBlueprintTaskCustomTaskMetadata_ = t.Dict(
    {
        t.Key("id"): String(allow_blank=False),
        t.Key("version_major"): Int(),
        t.Key("version_minor"): Int(),
        t.Key("label"): String(allow_blank=False),
    }
).allow_extra("*")

UserBlueprintTask_ = t.Dict(
    {
        t.Key("task_code"): String(allow_blank=False),
        t.Key("label"): String(allow_blank=True),
        t.Key("description"): String(allow_blank=True),
        t.Key("arguments"): t.List(UserBlueprintTaskArgument_),
        t.Key("categories"): t.List(String(allow_blank=False)),
        t.Key("colnames_and_types", optional=True): t.Or(t.List(ColnameAndType_), t.Null),
        t.Key("icon"): Int(),
        t.Key("output_methods"): t.List(String(allow_blank=False)),
        t.Key("time_series_only"): t.Bool(),
        t.Key("url"): t.Or(
            t.Dict().allow_extra("*"),
            String(allow_blank=True),
            TaskDocumentationUrl_,
        ),
        t.Key("valid_inputs"): t.List(String(allow_blank=False)),
        t.Key("is_custom_task", optional=True): t.Bool(),
        t.Key("custom_task_id", optional=True): t.Or(String, t.Null),
        t.Key("custom_task_versions", optional=True): t.List(UserBlueprintTaskCustomTaskMetadata_),
        t.Key("supports_scoring_code", optional=True): t.Bool(),
    }
).allow_extra("*")

UserBlueprintTaskLookupEntry_ = t.Dict(
    {
        t.Key("task_code"): String(allow_blank=False),
        t.Key("task_definition"): t.Or(UserBlueprintTask_),
    }
)

UserBlueprintTasksResponse_ = t.Dict(
    {
        t.Key("categories"): t.List(UserBlueprintTaskCategoryItem_),
        t.Key("tasks"): t.List(UserBlueprintTaskLookupEntry_),
    }
)

UserBlueprintSharedRolesResponseValidator_ = t.Dict(
    {
        t.Key("share_recipient_type"): t.Enum("user", "group", "organization"),
        t.Key("role"): t.Enum("CONSUMER", "EDITOR", "OWNER"),
        t.Key("id"): String(allow_blank=False, min_length=24, max_length=24),
        t.Key("name"): String(allow_blank=False),
    }
)

UserBlueprintSharedRolesListResponseValidator_ = t.Dict(
    {
        t.Key("count", optional=True): Int(),
        t.Key("next", optional=True): t.URL,
        t.Key("previous", optional=True): t.URL,
        t.Key("data"): t.List(UserBlueprintSharedRolesResponseValidator_),
        t.Key("total_count", optional=True): Int(),
    }
)


UserBlueprintCatalogSearchItem_ = t.Dict(
    {
        t.Key("id"): String(),
        t.Key("catalog_name"): String(),
        t.Key("description", optional=True): String(),
        t.Key("info_creator_full_name"): String(),
        t.Key("last_modifier_full_name", optional=True): String(),
        t.Key("user_blueprint_id"): String(),
    }
).allow_extra("*")


UserBlueprintValidationRequest_ = t.Dict(
    {
        t.Key("project_id", optional=True): t.String(allow_blank=False),
        t.Key("blueprint"): t.List(UserBlueprintsBlueprintTask_),
    }
)


UserBlueprintValidationResponse_ = t.Dict(
    {
        t.Key("vertex_context"): t.List(VertexContextItem_),
        t.Key("blueprint_context"): VertexContextItemMessages_,
    }
)
